# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A task consisting of finding goals/targets in a random maze."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import itertools

from dm_control import composer
from dm_control import mjcf
from dm_control.composer.observation import observable as observable_lib
from dm_control.locomotion.props import target_sphere
from dm_control.mujoco.wrapper import mjbindings
import numpy as np
from six.moves import range
from six.moves import zip

_NUM_RAYS = 10

# Aliveness in [-1., 0.].
DEFAULT_ALIVE_THRESHOLD = -0.5

DEFAULT_PHYSICS_TIMESTEP = 0.001
DEFAULT_CONTROL_TIMESTEP = 0.025


class NullGoalMaze(composer.Task):
  """A base task for maze with goals."""

  def __init__(self,
               walker,
               maze_arena,
               randomize_spawn_position=True,
               randomize_spawn_rotation=True,
               rotation_bias_factor=0,
               aliveness_reward=0.0,
               aliveness_threshold=DEFAULT_ALIVE_THRESHOLD,
               contact_termination=True,
               enable_global_task_observables=False,
               physics_timestep=DEFAULT_PHYSICS_TIMESTEP,
               control_timestep=DEFAULT_CONTROL_TIMESTEP):
    """Initializes goal-directed maze task.

    Args:
      walker: The body to navigate the maze.
      maze_arena: The physical maze arena object.
      randomize_spawn_position: Flag to randomize position of spawning.
      randomize_spawn_rotation: Flag to randomize orientation of spawning.
      rotation_bias_factor: A non-negative number that concentrates initial
        orientation away from walls. When set to zero, the initial orientation
        is uniformly random. The larger the value of this number, the more
        likely it is that the initial orientation would face the direction that
        is farthest away from a wall.
      aliveness_reward: Reward for being alive.
      aliveness_threshold: Threshold if should terminate based on walker
        aliveness feature.
      contact_termination: whether to terminate if a non-foot geom touches the
        ground.
      enable_global_task_observables: Flag to provide task observables that
        contain global information, including map layout.
      physics_timestep: timestep of simulation.
      control_timestep: timestep at which agent changes action.
    """
    self._walker = walker
    self._maze_arena = maze_arena
    self._walker.create_root_joints(self._maze_arena.attach(self._walker))

    self._randomize_spawn_position = randomize_spawn_position
    self._randomize_spawn_rotation = randomize_spawn_rotation
    self._rotation_bias_factor = rotation_bias_factor

    self._aliveness_reward = aliveness_reward
    self._aliveness_threshold = aliveness_threshold
    self._contact_termination = contact_termination
    self._discount = 1.0

    self.set_timesteps(
        physics_timestep=physics_timestep, control_timestep=control_timestep)

    self._walker.observables.egocentric_camera.height = 64
    self._walker.observables.egocentric_camera.width = 64

    for observable in (self._walker.observables.proprioception +
                       self._walker.observables.kinematic_sensors +
                       self._walker.observables.dynamic_sensors):
      observable.enabled = True
    self._walker.observables.egocentric_camera.enabled = True

    if enable_global_task_observables:
      # Reveal maze text map as observable.
      maze_obs = observable_lib.Generic(
          lambda _: self._maze_arena.maze.entity_layer)
      maze_obs.enabled = True

      # absolute walker position
      def get_walker_pos(physics):
        walker_pos = physics.bind(self._walker.root_body).xpos
        return walker_pos
      absolute_position = observable_lib.Generic(get_walker_pos)
      absolute_position.enabled = True

      # absolute walker orientation
      def get_walker_ori(physics):
        walker_ori = np.reshape(
            physics.bind(self._walker.root_body).xmat, (3, 3))
        return walker_ori
      absolute_orientation = observable_lib.Generic(get_walker_ori)
      absolute_orientation.enabled = True

      # grid element of player in maze cell: i,j cell in maze layout
      def get_walker_ij(physics):
        walker_xypos = physics.bind(self._walker.root_body).xpos[:-1]
        walker_rel_origin = (
            (walker_xypos +
             np.sign(walker_xypos) * self._maze_arena.xy_scale / 2) /
            (self._maze_arena.xy_scale)).astype(int)
        x_offset = (self._maze_arena.maze.width - 1) / 2
        y_offset = (self._maze_arena.maze.height - 1) / 2
        walker_ij = walker_rel_origin + np.array([x_offset, y_offset])
        return walker_ij
      absolute_position_discrete = observable_lib.Generic(get_walker_ij)
      absolute_position_discrete.enabled = True

      self._task_observables = collections.OrderedDict({
          'maze_layout': maze_obs,
          'absolute_position': absolute_position,
          'absolute_orientation': absolute_orientation,
          'location_in_maze': absolute_position_discrete,  # from bottom left
      })
    else:
      self._task_observables = collections.OrderedDict({})

  @property
  def task_observables(self):
    return self._task_observables

  @property
  def name(self):
    return 'goal_maze'

  @property
  def root_entity(self):
    return self._maze_arena

  def initialize_episode_mjcf(self, unused_random_state):
    self._maze_arena.regenerate()

  def _respawn(self, physics, random_state):
    self._walker.reinitialize_pose(physics, random_state)

    if self._randomize_spawn_position:
      self._spawn_position = self._maze_arena.spawn_positions[
          random_state.randint(0, len(self._maze_arena.spawn_positions))]

    if self._randomize_spawn_rotation:
      # Move walker up out of the way before raycasting.
      self._walker.shift_pose(physics, [0.0, 0.0, 100.0])

      distances = []
      geomid_out = np.array([-1], dtype=np.intc)
      for i in range(_NUM_RAYS):
        theta = 2 * np.pi * i / _NUM_RAYS
        pos = np.array([self._spawn_position[0], self._spawn_position[1], 0.1],
                       dtype=np.float64)
        vec = np.array([np.cos(theta), np.sin(theta), 0], dtype=np.float64)
        dist = mjbindings.mjlib.mj_ray(
            physics.model.ptr, physics.data.ptr, pos, vec,
            None, 1, -1, geomid_out)
        distances.append(dist)

      def remap_with_bias(x):
        """Remaps values [-1, 1] -> [-1, 1] with bias."""
        return np.tanh((1 + self._rotation_bias_factor) * np.arctanh(x))

      max_theta = 2 * np.pi * np.argmax(distances) / _NUM_RAYS
      rotation = max_theta + np.pi * (
          1 + remap_with_bias(random_state.uniform(-1, 1)))

      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]

      # Move walker back down.
      self._walker.shift_pose(physics, [0.0, 0.0, -100.0])
    else:
      quat = None

    self._walker.shift_pose(
        physics, [self._spawn_position[0], self._spawn_position[1], 0.0],
        quat,
        rotate_velocity=True)

  def initialize_episode(self, physics, random_state):
    super(NullGoalMaze, self).initialize_episode(physics, random_state)
    self._respawn(physics, random_state)
    self._discount = 1.0

    walker_foot_geoms = set(self._walker.ground_contact_geoms)
    walker_nonfoot_geoms = [
        geom for geom in self._walker.mjcf_model.find_all('geom')
        if geom not in walker_foot_geoms]
    self._walker_nonfoot_geomids = set(
        physics.bind(walker_nonfoot_geoms).element_id)
    self._ground_geomids = set(
        physics.bind(self._maze_arena.ground_geoms).element_id)

  def _is_disallowed_contact(self, contact):
    set1, set2 = self._walker_nonfoot_geomids, self._ground_geomids
    return ((contact.geom1 in set1 and contact.geom2 in set2) or
            (contact.geom1 in set2 and contact.geom2 in set1))

  def after_step(self, physics, random_state):
    self._failure_termination = False
    if self._contact_termination:
      for c in physics.data.contact:
        if self._is_disallowed_contact(c):
          self._failure_termination = True
          break

  def should_terminate_episode(self, physics):
    if self._walker.aliveness(physics) < self._aliveness_threshold:
      self._failure_termination = True
    if self._failure_termination:
      self._discount = 0.0
      return True
    else:
      return False

  def get_reward(self, physics):
    del physics
    return self._aliveness_reward

  def get_discount(self, physics):
    del physics
    return self._discount


class RepeatSingleGoalMaze(NullGoalMaze):
  """Requires an agent to repeatedly find the same goal in a maze."""

  def __init__(self,
               walker,
               maze_arena,
               target=None,
               target_reward_scale=1.0,
               randomize_spawn_position=True,
               randomize_spawn_rotation=True,
               rotation_bias_factor=0,
               aliveness_reward=0.0,
               aliveness_threshold=DEFAULT_ALIVE_THRESHOLD,
               contact_termination=True,
               max_repeats=0,
               enable_global_task_observables=False,
               physics_timestep=DEFAULT_PHYSICS_TIMESTEP,
               control_timestep=DEFAULT_CONTROL_TIMESTEP):
    super(RepeatSingleGoalMaze, self).__init__(
        walker=walker,
        maze_arena=maze_arena,
        randomize_spawn_position=randomize_spawn_position,
        randomize_spawn_rotation=randomize_spawn_rotation,
        rotation_bias_factor=rotation_bias_factor,
        aliveness_reward=aliveness_reward,
        aliveness_threshold=aliveness_threshold,
        contact_termination=contact_termination,
        enable_global_task_observables=enable_global_task_observables,
        physics_timestep=physics_timestep,
        control_timestep=control_timestep)
    if target is None:
      target = target_sphere.TargetSphere()
    self._target = target
    self._rewarded_this_step = False
    self._maze_arena.attach(target)
    self._target_reward_scale = target_reward_scale
    self._max_repeats = max_repeats
    self._targets_obtained = 0

    if enable_global_task_observables:
      xpos_origin_callable = lambda phys: phys.bind(walker.root_body).xpos

      def _target_pos(physics, target=target):
        return physics.bind(target.geom).xpos

      walker.observables.add_egocentric_vector(
          'target_0',
          observable_lib.Generic(_target_pos),
          origin_callable=xpos_origin_callable)

  def initialize_episode_mjcf(self, random_state):
    super(RepeatSingleGoalMaze, self).initialize_episode_mjcf(random_state)
    self._target_position = self._maze_arena.target_positions[
        random_state.randint(0, len(self._maze_arena.target_positions))]
    mjcf.get_attachment_frame(
        self._target.mjcf_model).pos = self._target_position

  def initialize_episode(self, physics, random_state):
    super(RepeatSingleGoalMaze, self).initialize_episode(physics, random_state)
    self._rewarded_this_step = False
    self._targets_obtained = 0

  def after_step(self, physics, random_state):
    super(RepeatSingleGoalMaze, self).after_step(physics, random_state)
    if self._target.activated:
      self._rewarded_this_step = True
      self._targets_obtained += 1
      if self._targets_obtained <= self._max_repeats:
        self._respawn(physics, random_state)
        self._target.reset(physics)
    else:
      self._rewarded_this_step = False

  def should_terminate_episode(self, physics):
    if super(RepeatSingleGoalMaze, self).should_terminate_episode(physics):
      return True
    if self._targets_obtained > self._max_repeats:
      return True

  def get_reward(self, physics):
    del physics
    if self._rewarded_this_step:
      target_reward = self._target_reward_scale
    else:
      target_reward = 0.0
    return target_reward + self._aliveness_reward


class ManyHeterogeneousGoalsMaze(NullGoalMaze):
  """Requires an agent to find multiple goals with different rewards."""

  def __init__(self,
               walker,
               maze_arena,
               target_builders,
               target_type_rewards,
               target_type_proportions,
               shuffle_target_builders=False,
               randomize_spawn_position=True,
               randomize_spawn_rotation=True,
               rotation_bias_factor=0,
               aliveness_reward=0.0,
               aliveness_threshold=DEFAULT_ALIVE_THRESHOLD,
               contact_termination=True,
               physics_timestep=DEFAULT_PHYSICS_TIMESTEP,
               control_timestep=DEFAULT_CONTROL_TIMESTEP):
    super(ManyHeterogeneousGoalsMaze, self).__init__(
        walker=walker,
        maze_arena=maze_arena,
        randomize_spawn_position=randomize_spawn_position,
        randomize_spawn_rotation=randomize_spawn_rotation,
        rotation_bias_factor=rotation_bias_factor,
        aliveness_reward=aliveness_reward,
        aliveness_threshold=aliveness_threshold,
        contact_termination=contact_termination,
        physics_timestep=physics_timestep,
        control_timestep=control_timestep)
    self._active_targets = []
    self._target_builders = target_builders
    self._target_type_rewards = tuple(target_type_rewards)
    self._target_type_fractions = (
        np.array(target_type_proportions, dtype=float) /
        np.sum(target_type_proportions))
    self._shuffle_target_builders = shuffle_target_builders

  def _get_targets(self, total_target_count, random_state):
    # Multiply total target count by the fraction for each type, rounded down.
    target_numbers = np.array([int(frac * total_target_count)
                               for frac in self._target_type_fractions])

    # Calculate deviations from the ideal ratio incurred by rounding.
    errors = (self._target_type_fractions -
              target_numbers / float(total_target_count))

    # Sort the target types by deviations from ideal ratios.
    target_types_sorted_by_errors = list(np.argsort(errors))

    # Top up individual target classes until we reach the desired total,
    # starting from the class that is furthest away from the ideal ratio.
    current_total = np.sum(target_numbers)
    while current_total < total_target_count:
      target_numbers[target_types_sorted_by_errors.pop()] += 1
      current_total += 1

    if self._shuffle_target_builders:
      random_state.shuffle(self._target_builders)

    all_targets = []
    for target_type, num in enumerate(target_numbers):
      targets = []
      target_builder = self._target_builders[target_type]
      for i in range(num):
        target = target_builder(name='target_{}_{}'.format(target_type, i))
        targets.append(target)
      all_targets.append(targets)
    return all_targets

  def initialize_episode_mjcf(self, random_state):
    super(
        ManyHeterogeneousGoalsMaze, self).initialize_episode_mjcf(random_state)
    for target in itertools.chain(*self._active_targets):
      target.detach()
    target_positions = list(self._maze_arena.target_positions)
    random_state.shuffle(target_positions)
    all_targets = self._get_targets(len(target_positions), random_state)
    for pos, target in zip(target_positions, itertools.chain(*all_targets)):
      self._maze_arena.attach(target)
      mjcf.get_attachment_frame(target.mjcf_model).pos = pos
      target.initialize_episode_mjcf(random_state)
    self._active_targets = all_targets
    self._target_rewarded = [[False] * len(targets) for targets in all_targets]

  def get_reward(self, physics):
    del physics
    reward = self._aliveness_reward
    for target_type, targets in enumerate(self._active_targets):
      for i, target in enumerate(targets):
        if target.activated and not self._target_rewarded[target_type][i]:
          reward += self._target_type_rewards[target_type]
          self._target_rewarded[target_type][i] = True
    return reward

  def should_terminate_episode(self, physics):
    if super(ManyHeterogeneousGoalsMaze,
             self).should_terminate_episode(physics):
      return True
    else:
      for target in itertools.chain(*self._active_targets):
        if not target.activated:
          return False
      # All targets have been activated: successful termination.
      return True


class ManyGoalsMaze(ManyHeterogeneousGoalsMaze):
  """Requires an agent to find all goals in a random maze."""

  def __init__(self,
               walker,
               maze_arena,
               target_builder,
               target_reward_scale=1.0,
               randomize_spawn_position=True,
               randomize_spawn_rotation=True,
               rotation_bias_factor=0,
               aliveness_reward=0.0,
               aliveness_threshold=DEFAULT_ALIVE_THRESHOLD,
               contact_termination=True,
               physics_timestep=DEFAULT_PHYSICS_TIMESTEP,
               control_timestep=DEFAULT_CONTROL_TIMESTEP):
    super(ManyGoalsMaze, self).__init__(
        walker=walker,
        maze_arena=maze_arena,
        target_builders=[target_builder],
        target_type_rewards=[target_reward_scale],
        target_type_proportions=[1],
        randomize_spawn_position=randomize_spawn_position,
        randomize_spawn_rotation=randomize_spawn_rotation,
        rotation_bias_factor=rotation_bias_factor,
        aliveness_reward=aliveness_reward,
        aliveness_threshold=aliveness_threshold,
        contact_termination=contact_termination,
        physics_timestep=physics_timestep,
        control_timestep=control_timestep)


class RepeatSingleGoalMazeAugmentedWithTargets(RepeatSingleGoalMaze):
  """Augments the single goal maze with many lower reward targets."""

  def __init__(self,
               walker,
               main_target,
               maze_arena,
               num_subtargets=20,
               target_reward_scale=10.0,
               subtarget_reward_scale=1.0,
               subtarget_colors=((0, 0, 0.4), (0, 0, 0.7)),
               randomize_spawn_position=True,
               randomize_spawn_rotation=True,
               rotation_bias_factor=0,
               aliveness_reward=0.0,
               aliveness_threshold=DEFAULT_ALIVE_THRESHOLD,
               contact_termination=True,
               physics_timestep=DEFAULT_PHYSICS_TIMESTEP,
               control_timestep=DEFAULT_CONTROL_TIMESTEP):
    super(RepeatSingleGoalMazeAugmentedWithTargets, self).__init__(
        walker=walker,
        target=main_target,
        maze_arena=maze_arena,
        target_reward_scale=target_reward_scale,
        randomize_spawn_position=randomize_spawn_position,
        randomize_spawn_rotation=randomize_spawn_rotation,
        rotation_bias_factor=rotation_bias_factor,
        aliveness_reward=aliveness_reward,
        aliveness_threshold=aliveness_threshold,
        contact_termination=contact_termination,
        physics_timestep=physics_timestep,
        control_timestep=control_timestep)
    self._subtarget_reward_scale = subtarget_reward_scale
    self._subtargets = []
    for i in range(num_subtargets):
      subtarget = target_sphere.TargetSphere(
          radius=0.4, rgb1=subtarget_colors[0], rgb2=subtarget_colors[1],
          name='subtarget_{}'.format(i)
      )
      self._subtargets.append(subtarget)
      self._maze_arena.attach(subtarget)
    self._subtarget_rewarded = None

  def initialize_episode_mjcf(self, random_state):
    super(RepeatSingleGoalMazeAugmentedWithTargets,
          self).initialize_episode_mjcf(random_state)
    subtarget_positions = self._maze_arena.target_positions
    for pos, subtarget in zip(subtarget_positions, self._subtargets):
      mjcf.get_attachment_frame(subtarget.mjcf_model).pos = pos
    self._subtarget_rewarded = [False] * len(self._subtargets)

  def get_reward(self, physics):
    main_reward = super(RepeatSingleGoalMazeAugmentedWithTargets,
                        self).get_reward(physics)
    subtarget_reward = 0
    for i, subtarget in enumerate(self._subtargets):
      if subtarget.activated and not self._subtarget_rewarded[i]:
        subtarget_reward += 1
        self._subtarget_rewarded[i] = True
    subtarget_reward *= self._subtarget_reward_scale
    return main_reward + subtarget_reward

  def should_terminate_episode(self, physics):
    if super(RepeatSingleGoalMazeAugmentedWithTargets,
             self).should_terminate_episode(physics):
      return True
    else:
      for subtarget in self._subtargets:
        if not subtarget.activated:
          return False
      # All subtargets have been activated.
      return True



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Utility classes."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import contextlib
import itertools
import sys
import time
import traceback

from absl import logging
import six

# Lower bound of the time multiplier set through TimeMultiplier class.
_MIN_TIME_MULTIPLIER = 1./16.
# Upper bound of the time multiplier set through TimeMultiplier class.
_MAX_TIME_MULTIPLIER = 64.


def is_scalar(value):
  """Checks if the supplied value can be converted to a scalar."""
  try:
    float(value)
  except (TypeError, ValueError):
    return False
  else:
    return True


def to_iterable(item):
  """Converts an item or iterable into an iterable."""
  if isinstance(item, six.string_types):
    return [item]
  elif isinstance(item, collections.Iterable):
    return item
  else:
    return [item]


class QuietSet(object):
  """A set-like container that quietly processes removals of missing keys."""

  def __init__(self):
    self._items = set()

  def __iadd__(self, items):
    """Adds `items`, avoiding duplicates.

    Args:
      items: An iterable of items to add, or a single item to add.

    Returns:
      This instance of `QuietSet`.
    """
    self._items.update(to_iterable(items))
    self._items.discard(None)
    return self

  def __isub__(self, items):
    """Detaches `items`.

    Args:
      items: An iterable of items to detach, or a single item to detach.

    Returns:
      This instance of `QuietSet`.
    """
    for item in to_iterable(items):
      self._items.discard(item)
    return self

  def __len__(self):
    return len(self._items)

  def __iter__(self):
    return iter(self._items)


def interleave(a, b):
  """Interleaves the contents of two iterables."""
  return itertools.chain.from_iterable(zip(a, b))


class TimeMultiplier(object):
  """Controls the relative speed of the simulation compared to realtime."""

  def __init__(self, initial_time_multiplier):
    """Instance initializer.

    Args:
      initial_time_multiplier: A float scalar specifying the initial speed of
        the simulation with 1.0 corresponding to realtime.
    """
    self.set(initial_time_multiplier)

  def get(self):
    """Returns the current time factor value."""
    return self._real_time_multiplier

  def set(self, value):
    """Modifies the time factor.

    Args:
      value: A float scalar, new value of the time factor.
    """
    self._real_time_multiplier = max(
        _MIN_TIME_MULTIPLIER, min(_MAX_TIME_MULTIPLIER, value))

  def __str__(self):
    """Returns a formatted string containing the time factor."""
    if self._real_time_multiplier >= 1.0:
      time_factor = '%d' % self._real_time_multiplier
    else:
      time_factor = '1/%d' % (1.0 // self._real_time_multiplier)
    return time_factor

  def increase(self):
    """Doubles the current time factor value."""
    self.set(self._real_time_multiplier * 2.)

  def decrease(self):
    """Halves the current time factor value."""
    self.set(self._real_time_multiplier / 2.)


class Integrator(object):
  """Integrates a value and averages it for the specified period of time."""

  def __init__(self, refresh_rate=.5):
    """Instance initializer.

    Args:
      refresh_rate: How often, in seconds, is the integrated value averaged.
    """
    self._value = 0
    self._value_acc = 0
    self._num_samples = 0
    self._sampling_timestamp = time.time()
    self._refresh_rate = refresh_rate

  @property
  def value(self):
    """Returns the averaged value."""
    return self._value

  @value.setter
  def value(self, val):
    """Integrates the new value."""
    self._value_acc += val
    self._num_samples += 1

    time_elapsed = time.time() - self._sampling_timestamp
    if time_elapsed >= self._refresh_rate:
      self._value = self._value_acc / self._num_samples
      self._value_acc = 0
      self._num_samples = 0
      self._sampling_timestamp = time.time()


class AtomicAction(object):
  """An action that cannot be interrupted."""

  def __init__(self, state_change_callback=None):
    """Instance initializer.

    Args:
      state_change_callback: Callable invoked when action changes its state.
    """
    self._state_change_callback = state_change_callback
    self._watermark = None

  def begin(self, watermark):
    """Begins the action, signing it with the specified watermark."""
    if self._watermark is None:
      self._watermark = watermark
      if self._state_change_callback is not None:
        self._state_change_callback(watermark)

  def end(self, watermark):
    """Ends a started action, provided the watermarks match."""
    if self._watermark == watermark:
      self._watermark = None
      if self._state_change_callback is not None:
        self._state_change_callback(None)

  @property
  def in_progress(self):
    """Returns a boolean value to indicate if the being method was called."""
    return self._watermark is not None

  @property
  def watermark(self):
    """Returns the watermark passed to begin() method call, or None.

    None will be returned if the action is not in progress.
    """
    return self._watermark


class ObservableFlag(QuietSet):
  """Observable boolean flag.

  The QuietState provides necessary functionality for managing listeners.

  A listener is a callable that takes one boolean parameter.
  """

  def __init__(self, initial_value):
    """Instance initializer.

    Args:
      initial_value: A boolean value with the initial state of the flag.
    """
    self._value = initial_value
    super(ObservableFlag, self).__init__()

  def toggle(self):
    """Toggles the value True/False."""
    self._value = not self._value
    for listener in self._items:
      listener(self._value)

  def __iadd__(self, value):
    """Add new listeners and update them about the state."""
    listeners = to_iterable(value)
    super(ObservableFlag, self).__iadd__(listeners)
    for listener in listeners:
      listener(self._value)
    return self

  @property
  def value(self):
    """Value of the flag."""
    return self._value

  @value.setter
  def value(self, val):
    if self._value != val:
      for listener in self._items:
        listener(self._value)
    self._value = val


class Timer(object):
  """Measures time elapsed between two ticks."""

  def __init__(self):
    """Instance initializer."""
    self._previous_time = time.time()
    self._measured_time = 0.

  def tick(self):
    """Updates the timer.

    Returns:
      Time elapsed since the last call to this method.
    """
    curr_time = time.time()
    self._measured_time = curr_time - self._previous_time
    self._previous_time = curr_time
    return self._measured_time

  @contextlib.contextmanager
  def measure_time(self):
    start_time = time.time()
    yield
    self._measured_time = time.time() - start_time

  @property
  def measured_time(self):
    return self._measured_time


class ErrorLogger(object):
  """A context manager that catches and logs all errors."""

  def __init__(self, listeners):
    """Instance initializer.

    Args:
      listeners: An iterable of callables, listeners to inform when an error
        is caught. Each callable should accept a single string argument.
    """
    self._error_found = False
    self._listeners = listeners

  def __enter__(self, *args):
    self._error_found = False

  def __exit__(self, exception_type, exception_value, tb):
    if exception_type:
      self._error_found = True
      error_message = ('dm_control viewer intercepted an environment error.\n'
                       'Original message: {}'.format(exception_value))
      logging.error(error_message)
      sys.stderr.write(error_message + '\nTraceback:\n')
      traceback.print_tb(tb)
      for listener in self._listeners:
        listener('{}'.format(exception_value))
      return True

  @property
  def errors_found(self):
    """Returns True if any errors were caught."""
    return self._error_found


class NullErrorLogger(object):
  """A context manager that replaces an ErrorLogger.

  This error logger will pass all thrown errors through.
  """

  def __enter__(self, *args):
    pass

  def __exit__(self, error_type, value, tb):
    pass

  @property
  def errors_found(self):
    """Returns True if any errors were caught."""
    return False



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Calculates a covering of text mazes with overlapping rectangular walls."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

import numpy as np
from six.moves import range

GridCoordinates = collections.namedtuple('GridCoordinates', ('y', 'x'))
MazeWall = collections.namedtuple('MazeWall', ('start', 'end'))


class _MazeWallCoveringContext(object):
  """Calculates a covering of text mazes with overlapping rectangular walls.

  This class uses a greedy algorithm to try and minimize the number of geoms
  generated to create a given maze. The solution is not guaranteed to be
  optimal, but in most cases should result in a significantly smaller number of
  geoms than if each cell were treated as an individual box.
  """

  def __init__(self, text_maze, wall_char='*', make_odd_sized_walls=False):
    """Initializes this _MazeWallCoveringContext.

    Args:
      text_maze: A `labmaze.TextGrid` instance.
      wall_char: (optional) The character that signifies a wall.
      make_odd_sized_walls: (optional) A boolean, if `True` all wall sections
        generated span odd numbers of grid cells. This option exists primarily
        to appease MuJoCo's texture repeating algorithm.
    """
    self._text_maze = text_maze
    self._wall_char = wall_char
    self._make_odd_sized_walls = make_odd_sized_walls
    self._covered = np.full(text_maze.shape, False, dtype=np.bool)
    self._maze_size = GridCoordinates(*text_maze.shape)
    self._next_start = GridCoordinates(0, 0)
    self._calculated = False
    self._walls = ()

  def calculate(self):
    """Calculates a covering of text mazes with overlapping rectangular walls.

    Returns:
      A tuple of `MazeWall` objects, each describing the corners of a wall.
    """
    if not self._calculated:
      self._calculated = True
      self._find_next_start()
      walls = []
      while self._next_start.y < self._maze_size.y:
        walls.append(self._find_next_wall())
        self._find_next_start()
      self._walls = tuple(walls)
    return self._walls

  def _find_next_start(self):
    """Moves `self._next_start` to the top-left corner of the next wall."""
    for y in range(self._next_start.y, self._maze_size.y):
      start_x = self._next_start.x if y == self._next_start.y else 0
      for x in range(start_x, self._maze_size.x):
        if self._text_maze[y, x] == self._wall_char and not self._covered[y, x]:
          self._next_start = GridCoordinates(y, x)
          return
    self._next_start = self._maze_size

  def _scan_row(self, row, start_col, end_col):
    """Scans a row of text maze to find the longest strip of wall."""
    for col in range(start_col, end_col):
      if (self._text_maze[row, col] != self._wall_char
          or self._covered[row, col]):
        return col
    return end_col

  def _find_next_wall(self):
    """Finds the largest piece of rectangular wall at the current location.

    This function assumes that `self._next_start` is already at the top-left
    corner of the next piece of wall.

    Returns:
      A `MazeWall` named tuple representing the next piece of wall created.
    """
    start = self._next_start
    x = self._maze_size.x
    end_x_for_rows = []
    total_cells = []

    for y in range(start.y, self._maze_size.y):
      x = self._scan_row(y, start.x, x)
      if x > start.x:
        if self._make_odd_sized_walls and (x - start.x) % 2 == 0:
          x -= 1
        end_x_for_rows.append(x)
        total_cells.append((x - start.x) * (y - start.y + 1))
        y += 1
      else:
        break

    if not self._make_odd_sized_walls:
      end_y_offset = total_cells.index(max(total_cells))
    else:
      end_y_offset = 2 * total_cells[::2].index(max(total_cells[::2]))
    end = GridCoordinates(start.y + end_y_offset + 1,
                          end_x_for_rows[end_y_offset])
    self._covered[start.y:end.y, start.x:end.x] = True
    self._next_start = GridCoordinates(start.y, end.x)
    return MazeWall(start, end)


def make_walls(text_maze, wall_char='*', make_odd_sized_walls=False):
  """Calculates a covering of text mazes with overlapping rectangular walls.

  Args:
    text_maze: A `labmaze.TextMaze` instance.
    wall_char: (optional) The character that signifies a wall.
    make_odd_sized_walls: (optional) A boolean, if `True` all wall sections
      generated span odd numbers of grid cells. This option exists primarily
      to appease MuJoCo's texture repeating algorithm.

  Returns:
    A tuple of `MazeWall` objects, each describing the corners of a wall.
  """
  wall_covering_context = _MazeWallCoveringContext(
      text_maze, wall_char=wall_char, make_odd_sized_walls=make_odd_sized_walls)
  return wall_covering_context.calculate()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Wrapper that adds pixel observations to a control environment."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

import dm_env
from dm_env import specs
import numpy as np

STATE_KEY = 'state'


class Wrapper(dm_env.Environment):
  """Wraps a control environment and adds an observation with profile data.

  The profile data describes the time Mujoco spent in a "step", and the
  observation consists of two values: the cumulative time spent on steps
  (in seconds), and the number of times the profiling timer was called.
  """

  def __init__(self, env, observation_key='step_timing'):
    """Initializes a new mujoco_profiling Wrapper.

    Args:
      env: The environment to wrap.
      observation_key: Optional custom string specifying the profile
        observation's key in the `OrderedDict` of observations. Defaults to
        'step_timing'.

    Raises:
      ValueError: If `env`'s observation spec is not compatible with the
        wrapper. Supported formats are a single array, or a dict of arrays.
      ValueError: If `env`'s observation already contains the specified
        `observation_key`.
    """
    wrapped_observation_spec = env.observation_spec()

    if isinstance(wrapped_observation_spec, specs.Array):
      self._observation_is_dict = False
      invalid_keys = set([STATE_KEY])
    elif isinstance(wrapped_observation_spec, collections.MutableMapping):
      self._observation_is_dict = True
      invalid_keys = set(wrapped_observation_spec.keys())
    else:
      raise ValueError('Unsupported observation spec structure.')

    if observation_key in invalid_keys:
      raise ValueError(
          'Duplicate or reserved observation key {!r}.'.format(observation_key))

    if self._observation_is_dict:
      self._observation_spec = wrapped_observation_spec.copy()
    else:
      self._observation_spec = collections.OrderedDict()
      self._observation_spec[STATE_KEY] = wrapped_observation_spec

    env.physics.enable_profiling()

    # Extend observation spec.
    self._observation_spec[observation_key] = specs.Array(
        shape=(2,), dtype=np.double, name=observation_key)

    self._env = env
    self._observation_key = observation_key

  def reset(self):
    return self._add_profile_observation(self._env.reset())

  def step(self, action):
    return self._add_profile_observation(self._env.step(action))

  def observation_spec(self):
    return self._observation_spec

  def action_spec(self):
    return self._env.action_spec()

  def _add_profile_observation(self, time_step):
    if self._observation_is_dict:
      observation = type(time_step.observation)(time_step.observation)
    else:
      observation = collections.OrderedDict()
      observation[STATE_KEY] = time_step.observation

    # timer[0] is the step timer. There are lots of different timers (see
    # mujoco/hdrs/mjdata.h)
    # but we only care about the step timer.
    timing = self._env.physics.data.timer[0]

    observation[self._observation_key] = np.array([timing[0], timing[1]],
                                                  dtype=np.double)
    return time_step._replace(observation=observation)

  def __getattr__(self, name):
    return getattr(self._env, name)



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for the user_input module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from dm_control.viewer import user_input
import mock


class InputMapTests(absltest.TestCase):

  def setUp(self):
    super(InputMapTests, self).setUp()
    self.mouse = mock.MagicMock()
    self.keyboard = mock.MagicMock()
    self.input_map = user_input.InputMap(self.mouse, self.keyboard)

    self.callback = mock.MagicMock()

  def test_clearing_bindings(self):
    self.input_map._active_exclusive = 1
    self.input_map._action_callbacks = {1: 2}
    self.input_map._double_click_callbacks = {3: 4}
    self.input_map._plane_callback = [5]
    self.input_map._z_axis_callback = [6]

    self.input_map.clear_bindings()

    self.assertEmpty(self.input_map._action_callbacks)
    self.assertEmpty(self.input_map._double_click_callbacks)
    self.assertEmpty(self.input_map._plane_callback)
    self.assertEmpty(self.input_map._z_axis_callback)
    self.assertEqual(
        user_input._NO_EXCLUSIVE_KEY, self.input_map._active_exclusive)

  def test_binding(self):
    self.input_map.bind(self.callback, user_input.KEY_UP)
    expected_dict = {
        (user_input.KEY_UP, user_input.MOD_NONE): (False, self.callback)}
    self.assertDictEqual(expected_dict, self.input_map._action_callbacks)

  def test_binding_exclusive(self):
    self.input_map.bind(self.callback, user_input.Exclusive(user_input.KEY_UP))
    expected_dict = {
        (user_input.KEY_UP, user_input.MOD_NONE): (True, self.callback)}
    self.assertDictEqual(expected_dict, self.input_map._action_callbacks)

  def test_binding_and_invoking_ranges_of_actions(self):
    self.input_map.bind(self.callback, user_input.Range(
        [user_input.KEY_UP, (user_input.KEY_UP, user_input.MOD_ALT)]))

    self.input_map._handle_key(
        user_input.KEY_UP, user_input.PRESS, user_input.MOD_NONE)
    self.callback.assert_called_once_with(0)

    self.callback.reset_mock()
    self.input_map._handle_key(
        user_input.KEY_UP, user_input.PRESS, user_input.MOD_ALT)
    self.callback.assert_called_once_with(1)

  def test_binding_planar_action(self):
    self.input_map.bind_plane(self.callback)
    self.assertLen(self.input_map._plane_callback, 1)
    self.assertEqual(self.callback, self.input_map._plane_callback[0])

  def test_binding_z_axis_action(self):
    self.input_map.bind_z_axis(self.callback)
    self.assertLen(self.input_map._z_axis_callback, 1)
    self.assertEqual(self.callback, self.input_map._z_axis_callback[0])

  def test_invoking_regular_action_in_response_to_click(self):
    self.input_map._action_callbacks = {(1, 2): (False, self.callback)}

    self.input_map._handle_key(1, user_input.PRESS, 2)
    self.callback.assert_called_once()
    self.callback.reset_mock()

    self.input_map._handle_key(1, user_input.RELEASE, 2)
    self.assertEqual(0, self.callback.call_count)

  def test_invoking_exclusive_action_in_response_to_click(self):
    self.input_map._action_callbacks = {(1, 2): (True, self.callback)}

    self.input_map._handle_key(1, user_input.PRESS, 2)
    self.callback.assert_called_once_with(True)
    self.callback.reset_mock()

    self.input_map._handle_key(1, user_input.RELEASE, 2)
    self.callback.assert_called_once_with(False)

  def test_exclusive_action_blocks_other_actions_until_its_finished(self):
    self.input_map._action_callbacks = {
        (1, 2): (True, self.callback), (3, 4): (False, self.callback)}

    self.input_map._handle_key(1, user_input.PRESS, 2)
    self.callback.assert_called_once_with(True)
    self.callback.reset_mock()

    # Attempting to start other actions (PRESS) or end them (RELEASE)
    # amounts to nothing.
    self.input_map._handle_key(3, user_input.PRESS, 4)
    self.assertEqual(0, self.callback.call_count)

    self.input_map._handle_key(3, user_input.RELEASE, 4)
    self.assertEqual(0, self.callback.call_count)

    # Even attempting to start the same action for the 2nd time fails.
    self.input_map._handle_key(1, user_input.PRESS, 2)
    self.assertEqual(0, self.callback.call_count)

    # Only finishing the action frees up the resources.
    self.input_map._handle_key(1, user_input.RELEASE, 2)
    self.callback.assert_called_once_with(False)
    self.callback.reset_mock()

    # Now we can start a new action.
    self.input_map._handle_key(3, user_input.PRESS, 4)
    self.callback.assert_called_once()

  def test_modifiers_required_only_for_exclusive_action_start(self):
    activation_modifiers = 2
    no_modifiers = 0
    self.input_map._action_callbacks = {
        (1, activation_modifiers): (True, self.callback)}

    self.input_map._handle_key(1, user_input.PRESS, activation_modifiers)
    self.callback.assert_called_once_with(True)
    self.callback.reset_mock()

    self.input_map._handle_key(1, user_input.RELEASE, no_modifiers)
    self.callback.assert_called_once_with(False)

  def test_invoking_regular_action_in_response_to_double_click(self):
    self.input_map._double_click_callbacks = {(1, 2): self.callback}
    self.input_map._handle_double_click(1, 2)
    self.callback.assert_called_once()

  def test_exclusive_actions_dont_respond_to_double_clicks(self):
    self.input_map._action_callbacks = {(1, 2): (True, self.callback)}

    self.input_map._handle_double_click(1, 2)
    self.assertEqual(0, self.callback.call_count)

  def test_mouse_move(self):
    position = [1, 2]
    translation = [3, 4]
    self.input_map._plane_callback = [self.callback]
    self.input_map._handle_mouse_move(position, translation)
    self.callback.assert_called_once_with(position, translation)

  def test_mouse_scroll(self):
    value = 5
    self.input_map._z_axis_callback = [self.callback]
    self.input_map._handle_mouse_scroll(value)
    self.callback.assert_called_once_with(value)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for locomotion.arenas.corridors."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized
from dm_control import mjcf
from dm_control.composer.variation import deterministic
from dm_control.locomotion.arenas import corridors
from six.moves import zip


class CorridorsTest(parameterized.TestCase):

  @parameterized.parameters([
      corridors.EmptyCorridor,
      corridors.GapsCorridor,
      corridors.WallsCorridor,
  ])
  def test_can_compile_mjcf(self, arena_type):
    arena = arena_type()
    mjcf.Physics.from_mjcf_model(arena.mjcf_model)

  @parameterized.parameters([
      corridors.EmptyCorridor,
      corridors.GapsCorridor,
      corridors.WallsCorridor,
  ])
  def test_can_regenerate_corridor_size(self, arena_type):
    width_sequence = [5.2, 3.8, 7.4]
    length_sequence = [21.1, 19.4, 16.3]

    arena = arena_type(
        corridor_width=deterministic.Sequence(width_sequence),
        corridor_length=deterministic.Sequence(length_sequence))

    # Add a probe geom that will generate contacts with the side walls.
    probe_body = arena.mjcf_model.worldbody.add('body', name='probe')
    probe_joint = probe_body.add('freejoint')
    probe_geom = probe_body.add('geom', name='probe', type='box')

    for expected_width, expected_length in zip(width_sequence, length_sequence):
      # No random_state is required since we are using deterministic variations.
      arena.regenerate(random_state=None)

      def resize_probe_geom_and_assert_num_contacts(
          delta_size, expected_num_contacts,
          expected_width=expected_width, expected_length=expected_length):
        probe_geom.size = [
            (expected_length / 2 + corridors._CORRIDOR_X_PADDING) + delta_size,
            expected_width / 2 + delta_size, 0.1]
        physics = mjcf.Physics.from_mjcf_model(arena.mjcf_model)
        probe_geomid = physics.bind(probe_geom).element_id
        physics.bind(probe_joint).qpos[:3] = [expected_length / 2, 0, 100]
        physics.forward()
        probe_contacts = [c for c in physics.data.contact
                          if c.geom1 == probe_geomid or c.geom2 == probe_geomid]
        self.assertLen(probe_contacts, expected_num_contacts)

      epsilon = 1e-7

      # If the probe geom is epsilon-smaller than the expected corridor size,
      # then we expect to detect no contact.
      resize_probe_geom_and_assert_num_contacts(-epsilon, 0)

      # If the probe geom is epsilon-larger than the expected corridor size,
      # then we expect to generate 4 contacts with each side wall, so 16 total.
      resize_probe_geom_and_assert_num_contacts(epsilon, 16)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for arenas.mazes.covering."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from dm_control.locomotion.arenas import covering
import labmaze
import numpy as np
import six
from six.moves import range

if six.PY3:
  _STRING_DTYPE = '|U1'
else:
  _STRING_DTYPE = '|S1'


class CoveringTest(absltest.TestCase):

  def testRandomMazes(self):
    maze = labmaze.RandomMaze(height=17, width=17,
                              max_rooms=5, room_min_size=3, room_max_size=5,
                              spawns_per_room=0, objects_per_room=0,
                              random_seed=54321)
    for _ in range(1000):
      maze.regenerate()
      walls = covering.make_walls(maze.entity_layer)
      reconstructed = np.full(maze.entity_layer.shape, ' ', dtype=_STRING_DTYPE)
      for wall in walls:
        reconstructed[wall.start.y:wall.end.y, wall.start.x:wall.end.x] = '*'
      np.testing.assert_array_equal(reconstructed, maze.entity_layer)

  def testOddCovering(self):
    maze = labmaze.RandomMaze(height=17, width=17,
                              max_rooms=5, room_min_size=3, room_max_size=5,
                              spawns_per_room=0, objects_per_room=0,
                              random_seed=54321)
    for _ in range(1000):
      maze.regenerate()
      walls = covering.make_walls(maze.entity_layer, make_odd_sized_walls=True)
      reconstructed = np.full(maze.entity_layer.shape, ' ', dtype=_STRING_DTYPE)
      for wall in walls:
        reconstructed[wall.start.y:wall.end.y, wall.start.x:wall.end.x] = '*'
      np.testing.assert_array_equal(reconstructed, maze.entity_layer)
      for wall in walls:
        self.assertEqual((wall.end.y - wall.start.y) % 2, 1)
        self.assertEqual((wall.end.x - wall.start.x) % 2, 1)

  def testNoOverlappingWalls(self):
    maze_string = """..**
                     .***
                     .***
                     """.replace(' ', '')
    walls = covering.make_walls(labmaze.TextGrid(maze_string))
    surface = 0
    for wall in walls:
      size_x = wall.end.x - wall.start.x
      size_y = wall.end.y - wall.start.y
      surface += size_x * size_y
    self.assertEqual(surface, 8)


if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests to check whether methods of `mujoco.Physics` are threadsafe."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import platform

# Internal dependencies.

from absl.testing import absltest
from dm_control import _render
from dm_control.mujoco import engine
from dm_control.mujoco.testing import assets
from dm_control.mujoco.testing import decorators

from six.moves import range

MODEL = assets.get_contents('cartpole.xml')
NUM_STEPS = 10

# Context creation with GLFW is not threadsafe.
if _render.BACKEND == 'glfw':
  # On Linux we are able to create a GLFW window in a single thread that is not
  # the main thread.
  # On Mac we are only allowed to create windows on the main thread, so we
  # disable the `run_threaded` wrapper entirely.
  NUM_THREADS = None if platform.system() == 'Darwin' else 1
else:
  NUM_THREADS = 4


class ThreadSafetyTest(absltest.TestCase):

  @decorators.run_threaded(num_threads=NUM_THREADS)
  def test_load_physics_from_string(self):
    engine.Physics.from_xml_string(MODEL)

  @decorators.run_threaded(num_threads=NUM_THREADS)
  def test_load_and_reload_physics_from_string(self):
    physics = engine.Physics.from_xml_string(MODEL)
    physics.reload_from_xml_string(MODEL)

  @decorators.run_threaded(num_threads=NUM_THREADS)
  def test_load_and_step_physics(self):
    physics = engine.Physics.from_xml_string(MODEL)
    for _ in range(NUM_STEPS):
      physics.step()

  @decorators.run_threaded(num_threads=NUM_THREADS)
  def test_load_and_step_multiple_physics_parallel(self):
    physics1 = engine.Physics.from_xml_string(MODEL)
    physics2 = engine.Physics.from_xml_string(MODEL)
    for _ in range(NUM_STEPS):
      physics1.step()
      physics2.step()

  @decorators.run_threaded(num_threads=NUM_THREADS)
  def test_load_and_step_multiple_physics_sequential(self):
    physics1 = engine.Physics.from_xml_string(MODEL)
    for _ in range(NUM_STEPS):
      physics1.step()
    del physics1
    physics2 = engine.Physics.from_xml_string(MODEL)
    for _ in range(NUM_STEPS):
      physics2.step()

  @decorators.run_threaded(num_threads=NUM_THREADS, calls_per_thread=5)
  def test_load_physics_and_render(self):
    physics = engine.Physics.from_xml_string(MODEL)

    # Check that frames aren't repeated - make the cartpole move.
    physics.set_control([1.0])

    unique_frames = set()
    for _ in range(NUM_STEPS):
      physics.step()
      frame = physics.render(width=320, height=240, camera_id=0)
      unique_frames.add(frame.tostring())

    self.assertLen(unique_frames, NUM_STEPS)

  @decorators.run_threaded(num_threads=NUM_THREADS, calls_per_thread=5)
  def test_render_multiple_physics_instances_per_thread_parallel(self):
    physics1 = engine.Physics.from_xml_string(MODEL)
    physics2 = engine.Physics.from_xml_string(MODEL)
    for _ in range(NUM_STEPS):
      physics1.step()
      physics1.render(width=320, height=240, camera_id=0)
      physics2.step()
      physics2.render(width=320, height=240, camera_id=0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Planar Walker Domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from dm_control.suite import base
from dm_control.suite import common
from dm_control.suite.utils import randomizers
from dm_control.utils import containers
from dm_control.utils import rewards


_DEFAULT_TIME_LIMIT = 25
_CONTROL_TIMESTEP = .025

# Minimal height of torso over foot above which stand reward is 1.
_STAND_HEIGHT = 1.2

# Horizontal speeds (meters/second) above which move reward is 1.
_WALK_SPEED = 1
_RUN_SPEED = 8


SUITE = containers.TaggedTasks()


def get_model_and_assets():
  """Returns a tuple containing the model XML string and a dict of assets."""
  return common.read_model('walker.xml'), common.ASSETS


@SUITE.add('benchmarking')
def stand(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns the Stand task."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = PlanarWalker(move_speed=0, random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, time_limit=time_limit, control_timestep=_CONTROL_TIMESTEP,
      **environment_kwargs)


@SUITE.add('benchmarking')
def walk(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns the Walk task."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = PlanarWalker(move_speed=_WALK_SPEED, random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, time_limit=time_limit, control_timestep=_CONTROL_TIMESTEP,
      **environment_kwargs)


@SUITE.add('benchmarking')
def run(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns the Run task."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = PlanarWalker(move_speed=_RUN_SPEED, random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, time_limit=time_limit, control_timestep=_CONTROL_TIMESTEP,
      **environment_kwargs)


class Physics(mujoco.Physics):
  """Physics simulation with additional features for the Walker domain."""

  def torso_upright(self):
    """Returns projection from z-axes of torso to the z-axes of world."""
    return self.named.data.xmat['torso', 'zz']

  def torso_height(self):
    """Returns the height of the torso."""
    return self.named.data.xpos['torso', 'z']

  def horizontal_velocity(self):
    """Returns the horizontal velocity of the center-of-mass."""
    return self.named.data.sensordata['torso_subtreelinvel'][0]

  def orientations(self):
    """Returns planar orientations of all bodies."""
    return self.named.data.xmat[1:, ['xx', 'xz']].ravel()


class PlanarWalker(base.Task):
  """A planar walker task."""

  def __init__(self, move_speed, random=None):
    """Initializes an instance of `PlanarWalker`.

    Args:
      move_speed: A float. If this value is zero, reward is given simply for
        standing up. Otherwise this specifies a target horizontal velocity for
        the walking task.
      random: Optional, either a `numpy.random.RandomState` instance, an
        integer seed for creating a new `RandomState`, or None to select a seed
        automatically (default).
    """
    self._move_speed = move_speed
    super(PlanarWalker, self).__init__(random=random)

  def initialize_episode(self, physics):
    """Sets the state of the environment at the start of each episode.

    In 'standing' mode, use initial orientation and small velocities.
    In 'random' mode, randomize joint angles and let fall to the floor.

    Args:
      physics: An instance of `Physics`.

    """
    randomizers.randomize_limited_and_rotational_joints(physics, self.random)
    super(PlanarWalker, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns an observation of body orientations, height and velocites."""
    obs = collections.OrderedDict()
    obs['orientations'] = physics.orientations()
    obs['height'] = physics.torso_height()
    obs['velocity'] = physics.velocity()
    return obs

  def get_reward(self, physics):
    """Returns a reward to the agent."""
    standing = rewards.tolerance(physics.torso_height(),
                                 bounds=(_STAND_HEIGHT, float('inf')),
                                 margin=_STAND_HEIGHT/2)
    upright = (1 + physics.torso_upright()) / 2
    stand_reward = (3*standing + upright) / 4
    if self._move_speed == 0:
      return stand_reward
    else:
      move_reward = rewards.tolerance(physics.horizontal_velocity(),
                                      bounds=(self._move_speed, float('inf')),
                                      margin=self._move_speed/2,
                                      value_at_margin=0.5,
                                      sigmoid='linear')
      return stand_reward * (5*move_reward + 1) / 6



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Wrapper that adds pixel observations to a control environment."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

import dm_env
from dm_env import specs

STATE_KEY = 'state'


class Wrapper(dm_env.Environment):
  """Wraps a control environment and adds a rendered pixel observation."""

  def __init__(self, env, pixels_only=True, render_kwargs=None,
               observation_key='pixels'):
    """Initializes a new pixel Wrapper.

    Args:
      env: The environment to wrap.
      pixels_only: If True (default), the original set of 'state' observations
        returned by the wrapped environment will be discarded, and the
        `OrderedDict` of observations will only contain pixels. If False, the
        `OrderedDict` will contain the original observations as well as the
        pixel observations.
      render_kwargs: Optional `dict` containing keyword arguments passed to the
        `mujoco.Physics.render` method.
      observation_key: Optional custom string specifying the pixel observation's
        key in the `OrderedDict` of observations. Defaults to 'pixels'.

    Raises:
      ValueError: If `env`'s observation spec is not compatible with the
        wrapper. Supported formats are a single array, or a dict of arrays.
      ValueError: If `env`'s observation already contains the specified
        `observation_key`.
    """
    if render_kwargs is None:
      render_kwargs = {}

    wrapped_observation_spec = env.observation_spec()

    if isinstance(wrapped_observation_spec, specs.Array):
      self._observation_is_dict = False
      invalid_keys = set([STATE_KEY])
    elif isinstance(wrapped_observation_spec, collections.MutableMapping):
      self._observation_is_dict = True
      invalid_keys = set(wrapped_observation_spec.keys())
    else:
      raise ValueError('Unsupported observation spec structure.')

    if not pixels_only and observation_key in invalid_keys:
      raise ValueError('Duplicate or reserved observation key {!r}.'
                       .format(observation_key))

    if pixels_only:
      self._observation_spec = collections.OrderedDict()
    elif self._observation_is_dict:
      self._observation_spec = wrapped_observation_spec.copy()
    else:
      self._observation_spec = collections.OrderedDict()
      self._observation_spec[STATE_KEY] = wrapped_observation_spec

    # Extend observation spec.
    pixels = env.physics.render(**render_kwargs)
    pixels_spec = specs.Array(
        shape=pixels.shape, dtype=pixels.dtype, name=observation_key)
    self._observation_spec[observation_key] = pixels_spec

    self._env = env
    self._pixels_only = pixels_only
    self._render_kwargs = render_kwargs
    self._observation_key = observation_key

  def reset(self):
    time_step = self._env.reset()
    return self._add_pixel_observation(time_step)

  def step(self, action):
    time_step = self._env.step(action)
    return self._add_pixel_observation(time_step)

  def observation_spec(self):
    return self._observation_spec

  def action_spec(self):
    return self._env.action_spec()

  def _add_pixel_observation(self, time_step):
    if self._pixels_only:
      observation = collections.OrderedDict()
    elif self._observation_is_dict:
      observation = type(time_step.observation)(time_step.observation)
    else:
      observation = collections.OrderedDict()
      observation[STATE_KEY] = time_step.observation

    pixels = self._env.physics.render(**self._render_kwargs)
    observation[self._observation_key] = pixels
    return time_step._replace(observation=observation)

  def __getattr__(self, name):
    return getattr(self._env, name)



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for locomotion.tasks.soccer."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
# Internal dependencies.

from absl.testing import absltest
from absl.testing import parameterized
from dm_control import composer
from dm_control import mjcf
from dm_control.locomotion import soccer
from dm_control.locomotion.soccer import camera
from dm_control.locomotion.soccer import initializers
from dm_control.mujoco.wrapper import mjbindings
import numpy as np
from six.moves import range
from six.moves import zip

RGBA_BLUE = [.1, .1, .8, 1.]
RGBA_RED = [.8, .1, .1, 1.]


def _walker(name, walker_id, marker_rgba):
  return soccer.BoxHead(
      name=name,
      walker_id=walker_id,
      marker_rgba=marker_rgba,
  )


def _team_players(team_size, team, team_name, team_color):
  team_of_players = []
  for i in range(team_size):
    team_of_players.append(
        soccer.Player(team, _walker("%s%d" % (team_name, i), i, team_color)))
  return team_of_players


def _home_team(team_size):
  return _team_players(team_size, soccer.Team.HOME, "home", RGBA_BLUE)


def _away_team(team_size):
  return _team_players(team_size, soccer.Team.AWAY, "away", RGBA_RED)


def _env(players, disable_walker_contacts=True, observables=None,
         random_state=42, **task_kwargs):
  return composer.Environment(
      task=soccer.Task(
          players=players,
          arena=soccer.Pitch((20, 15)),
          observables=observables,
          disable_walker_contacts=disable_walker_contacts,
          **task_kwargs
      ),
      random_state=random_state,
      time_limit=1)


def _observables_adder(observables_adder):
  if observables_adder == "core":
    return soccer.CoreObservablesAdder()
  if observables_adder == "core_interception":
    return soccer.MultiObservablesAdder(
        [soccer.CoreObservablesAdder(),
         soccer.InterceptionObservablesAdder()])
  raise ValueError("Unrecognized observable_adder %s" % observables_adder)


class TaskTest(parameterized.TestCase):

  def _assert_all_count_equal(self, list_of_lists):
    """Check all lists in the list are count equal."""
    if not list_of_lists:
      return

    first = sorted(list_of_lists[0])
    for other in list_of_lists[1:]:
      self.assertCountEqual(first, other)

  @parameterized.named_parameters(
      ("1vs1_core", 1, "core", 33, True),
      ("2vs2_core", 2, "core", 43, True),
      ("1vs1_interception", 1, "core_interception", 41, True),
      ("2vs2_interception", 2, "core_interception", 51, True),
      ("1vs1_core_contact", 1, "core", 33, False),
      ("2vs2_core_contact", 2, "core", 43, False),
      ("1vs1_interception_contact", 1, "core_interception", 41, False),
      ("2vs2_interception_contact", 2, "core_interception", 51, False),
  )
  def test_step_environment(self, team_size, observables_adder, num_obs,
                            disable_walker_contacts):
    env = _env(
        _home_team(team_size) + _away_team(team_size),
        observables=_observables_adder(observables_adder),
        disable_walker_contacts=disable_walker_contacts)
    self.assertLen(env.action_spec(), 2 * team_size)
    self.assertLen(env.observation_spec(), 2 * team_size)

    actions = [np.zeros(s.shape, s.dtype) for s in env.action_spec()]

    timestep = env.reset()

    for observation, spec in zip(timestep.observation, env.observation_spec()):
      self.assertLen(spec, num_obs)
      self.assertCountEqual(list(observation.keys()), list(spec.keys()))
      for key in observation.keys():
        self.assertEqual(observation[key].shape, spec[key].shape)

    while not timestep.last():
      timestep = env.step(actions)

  # TODO(b/124848293): consolidate environment stepping loop for task tests.
  @parameterized.named_parameters(
      ("1vs2", 1, 2, 38),
      ("2vs1", 2, 1, 38),
      ("3vs0", 3, 0, 38),
      ("0vs2", 0, 2, 33),
      ("2vs2", 2, 2, 43),
      ("0vs0", 0, 0, None),
  )
  def test_num_players(self, home_size, away_size, num_observations):
    env = _env(_home_team(home_size) + _away_team(away_size))
    self.assertLen(env.action_spec(), home_size + away_size)
    self.assertLen(env.observation_spec(), home_size + away_size)

    actions = [np.zeros(s.shape, s.dtype) for s in env.action_spec()]

    timestep = env.reset()

    # Members of the same team should have identical specs.
    self._assert_all_count_equal(
        [spec.keys() for spec in env.observation_spec()[:home_size]])
    self._assert_all_count_equal(
        [spec.keys() for spec in env.observation_spec()[-away_size:]])

    for observation, spec in zip(timestep.observation, env.observation_spec()):
      self.assertCountEqual(list(observation.keys()), list(spec.keys()))
      for key in observation.keys():
        self.assertEqual(observation[key].shape, spec[key].shape)

      self.assertLen(spec, num_observations)

    while not timestep.last():
      timestep = env.step(actions)

      self.assertLen(timestep.observation, home_size + away_size)

      self.assertLen(timestep.reward, home_size + away_size)
      for player_spec, player_reward in zip(env.reward_spec(), timestep.reward):
        player_spec.validate(player_reward)

      discount_spec = env.discount_spec()
      discount_spec.validate(timestep.discount)

  def test_all_contacts(self):
    env = _env(_home_team(1) + _away_team(1))

    def _all_contact_configuration(physics, unused_random_state):
      walkers = [p.walker for p in env.task.players]
      ball = env.task.ball

      x, y, rotation = 0., 0., np.pi / 6.
      ball.set_pose(physics, [x, y, 5.])
      ball.set_velocity(
          physics, velocity=np.zeros(3), angular_velocity=np.zeros(3))

      x, y, rotation = 0., 0., np.pi / 3.
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[0].set_pose(physics, [x, y, 3.], quat)
      walkers[0].set_velocity(
          physics, velocity=np.zeros(3), angular_velocity=np.zeros(3))

      x, y, rotation = 0., 0., np.pi / 3. + np.pi
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[1].set_pose(physics, [x, y, 1.], quat)
      walkers[1].set_velocity(
          physics, velocity=np.zeros(3), angular_velocity=np.zeros(3))

    env.add_extra_hook("initialize_episode", _all_contact_configuration)

    actions = [np.zeros(s.shape, s.dtype) for s in env.action_spec()]

    timestep = env.reset()
    while not timestep.last():
      timestep = env.step(actions)

  def test_symmetric_observations(self):
    env = _env(_home_team(1) + _away_team(1))

    def _symmetric_configuration(physics, unused_random_state):
      walkers = [p.walker for p in env.task.players]
      ball = env.task.ball

      x, y, rotation = 0., 0., np.pi / 6.
      ball.set_pose(physics, [x, y, 0.5])
      ball.set_velocity(
          physics, velocity=np.zeros(3), angular_velocity=np.zeros(3))

      x, y, rotation = 5., 3., np.pi / 3.
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[0].set_pose(physics, [x, y, 0.], quat)
      walkers[0].set_velocity(
          physics, velocity=np.zeros(3), angular_velocity=np.zeros(3))

      x, y, rotation = -5., -3., np.pi / 3. + np.pi
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[1].set_pose(physics, [x, y, 0.], quat)
      walkers[1].set_velocity(
          physics, velocity=np.zeros(3), angular_velocity=np.zeros(3))

    env.add_extra_hook("initialize_episode", _symmetric_configuration)

    timestep = env.reset()
    obs_a, obs_b = timestep.observation
    self.assertCountEqual(list(obs_a.keys()), list(obs_b.keys()))
    for k in sorted(obs_a.keys()):
      o_a, o_b = obs_a[k], obs_b[k]
      np.testing.assert_allclose(o_a, o_b, err_msg=k + " not equal.", atol=1e-6)

  def test_symmetric_dynamic_observations(self):
    env = _env(_home_team(1) + _away_team(1))

    def _symmetric_configuration(physics, unused_random_state):
      walkers = [p.walker for p in env.task.players]
      ball = env.task.ball

      x, y, rotation = 0., 0., np.pi / 6.
      ball.set_pose(physics, [x, y, 0.5])
      # Ball shooting up. Walkers going tangent.
      ball.set_velocity(physics, velocity=[0., 0., 1.],
                        angular_velocity=[0., 0., 0.])

      x, y, rotation = 5., 3., np.pi / 3.
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[0].set_pose(physics, [x, y, 0.], quat)
      walkers[0].set_velocity(physics, velocity=[y, -x, 0.],
                              angular_velocity=[0., 0., 0.])

      x, y, rotation = -5., -3., np.pi / 3. + np.pi
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[1].set_pose(physics, [x, y, 0.], quat)
      walkers[1].set_velocity(physics, velocity=[y, -x, 0.],
                              angular_velocity=[0., 0., 0.])

    env.add_extra_hook("initialize_episode", _symmetric_configuration)

    timestep = env.reset()
    obs_a, obs_b = timestep.observation
    self.assertCountEqual(list(obs_a.keys()), list(obs_b.keys()))
    for k in sorted(obs_a.keys()):
      o_a, o_b = obs_a[k], obs_b[k]
      np.testing.assert_allclose(o_a, o_b, err_msg=k + " not equal.", atol=1e-6)

  def test_prev_actions(self):
    env = _env(_home_team(1) + _away_team(1))

    actions = []
    for i, player in enumerate(env.task.players):
      spec = player.walker.action_spec
      actions.append((i + 1) * np.ones(spec.shape, dtype=spec.dtype))

    env.reset()
    timestep = env.step(actions)

    for walker_idx, obs in enumerate(timestep.observation):
      np.testing.assert_allclose(
          np.squeeze(obs["prev_action"], axis=0),
          actions[walker_idx],
          err_msg="Walker {}: incorrect previous action.".format(walker_idx))

  @parameterized.named_parameters(
      dict(testcase_name="1vs2_draw",
           home_size=1, away_size=2, ball_vel_x=0, expected_home_score=0),
      dict(testcase_name="1vs2_home_score",
           home_size=1, away_size=2, ball_vel_x=50, expected_home_score=1),
      dict(testcase_name="2vs1_away_score",
           home_size=2, away_size=1, ball_vel_x=-50, expected_home_score=-1),
      dict(testcase_name="3vs0_home_score",
           home_size=3, away_size=0, ball_vel_x=50, expected_home_score=1),
      dict(testcase_name="0vs2_home_score",
           home_size=0, away_size=2, ball_vel_x=50, expected_home_score=1),
      dict(testcase_name="2vs2_away_score",
           home_size=2, away_size=2, ball_vel_x=-50, expected_home_score=-1),
  )
  def test_scoring_rewards(
      self, home_size, away_size, ball_vel_x, expected_home_score):
    env = _env(_home_team(home_size) + _away_team(away_size))

    def _score_configuration(physics, random_state):
      del random_state  # Unused.
      # Send the ball shooting towards either the home or away goal.
      env.task.ball.set_pose(physics, [0., 0., 0.5])
      env.task.ball.set_velocity(physics,
                                 velocity=[ball_vel_x, 0., 0.],
                                 angular_velocity=[0., 0., 0.])

    env.add_extra_hook("initialize_episode", _score_configuration)

    actions = [np.zeros(s.shape, s.dtype) for s in env.action_spec()]

    # Disable contacts and gravity so that the ball follows a straight path.
    with env.physics.model.disable("contact", "gravity"):

      timestep = env.reset()
      with self.subTest("Reward and discount are None on the first timestep"):
        self.assertTrue(timestep.first())
        self.assertIsNone(timestep.reward)
        self.assertIsNone(timestep.discount)

      # Step until the episode ends.
      timestep = env.step(actions)
      while not timestep.last():
        self.assertTrue(timestep.mid())
        # For non-terminal timesteps, the reward should always be 0 and the
        # discount should always be 1.
        np.testing.assert_array_equal(np.hstack(timestep.reward), 0.)
        self.assertEqual(timestep.discount, 1.)
        timestep = env.step(actions)

    # If a goal was scored then the epsiode should have ended with a discount of
    # 0. If neither team scored and the episode ended due to hitting the time
    # limit then the discount should be 1.
    with self.subTest("Correct terminal discount"):
      if expected_home_score != 0:
        expected_discount = 0.
      else:
        expected_discount = 1.
      self.assertEqual(timestep.discount, expected_discount)

    with self.subTest("Correct terminal reward"):
      reward = np.hstack(timestep.reward)
      np.testing.assert_array_equal(reward[:home_size], expected_home_score)
      np.testing.assert_array_equal(reward[home_size:], -expected_home_score)

  def test_throw_in(self):
    env = _env(_home_team(1) + _away_team(1))

    def _throw_in_configuration(physics, unused_random_state):
      walkers = [p.walker for p in env.task.players]
      ball = env.task.ball

      x, y, rotation = 0., 3., np.pi / 6.
      ball.set_pose(physics, [x, y, 0.5])
      # Ball shooting out of bounds.
      ball.set_velocity(physics, velocity=[0., 50., 0.],
                        angular_velocity=[0., 0., 0.])

      x, y, rotation = 0., -3., np.pi / 3.
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[0].set_pose(physics, [x, y, 0.], quat)
      walkers[0].set_velocity(physics, velocity=[0., 0., 0.],
                              angular_velocity=[0., 0., 0.])
      x, y, rotation = 0., -5., np.pi / 3.
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[1].set_pose(physics, [x, y, 0.], quat)
      walkers[1].set_velocity(physics, velocity=[0., 0., 0.],
                              angular_velocity=[0., 0., 0.])

    env.add_extra_hook("initialize_episode", _throw_in_configuration)

    actions = [np.zeros(s.shape, s.dtype) for s in env.action_spec()]

    timestep = env.reset()

    while not timestep.last():
      timestep = env.step(actions)

    terminal_ball_vel = np.linalg.norm(
        timestep.observation[0]["ball_ego_linear_velocity"])
    self.assertAlmostEqual(terminal_ball_vel, 0.)

  @parameterized.named_parameters(("score", 50., 0.), ("timeout", 0., 1.))
  def test_terminal_discount(self, init_ball_vel_x, expected_terminal_discount):
    env = _env(_home_team(1) + _away_team(1))

    def _initial_configuration(physics, unused_random_state):
      walkers = [p.walker for p in env.task.players]
      ball = env.task.ball

      x, y, rotation = 0., 0., np.pi / 6.
      ball.set_pose(physics, [x, y, 0.5])
      # Ball shooting up. Walkers going tangent.
      ball.set_velocity(physics, velocity=[init_ball_vel_x, 0., 0.],
                        angular_velocity=[0., 0., 0.])

      x, y, rotation = 0., -3., np.pi / 3.
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[0].set_pose(physics, [x, y, 0.], quat)
      walkers[0].set_velocity(physics, velocity=[0., 0., 0.],
                              angular_velocity=[0., 0., 0.])
      x, y, rotation = 0., 3., np.pi / 3.
      quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
      walkers[1].set_pose(physics, [x, y, 0.], quat)
      walkers[1].set_velocity(physics, velocity=[0., 0., 0.],
                              angular_velocity=[0., 0., 0.])

    env.add_extra_hook("initialize_episode", _initial_configuration)

    actions = [np.zeros(s.shape, s.dtype) for s in env.action_spec()]

    timestep = env.reset()

    while not timestep.last():
      timestep = env.step(actions)

    self.assertEqual(timestep.discount, expected_terminal_discount)

  @parameterized.named_parameters(("reset_only", False), ("step", True))
  def test_render(self, take_step):
    height = 100
    width = 150
    tracking_cameras = []
    for min_distance in [1, 1, 2]:
      tracking_cameras.append(
          camera.MultiplayerTrackingCamera(
              min_distance=min_distance,
              distance_factor=1,
              smoothing_update_speed=0.1,
              width=width,
              height=height,
          ))
    env = _env(_home_team(1) + _away_team(1), tracking_cameras=tracking_cameras)
    env.reset()
    if take_step:
      actions = [np.zeros(s.shape, s.dtype) for s in env.action_spec()]
      env.step(actions)
    rendered_frames = [cam.render() for cam in tracking_cameras]
    for frame in rendered_frames:
      assert frame.shape == (height, width, 3)
    self.assertTrue(np.array_equal(rendered_frames[0], rendered_frames[1]))
    self.assertFalse(np.array_equal(rendered_frames[1], rendered_frames[2]))


class UniformInitializerTest(parameterized.TestCase):

  @parameterized.parameters([0.3, 0.7])
  def test_walker_position(self, spawn_ratio):
    initializer = initializers.UniformInitializer(spawn_ratio=spawn_ratio)
    env = _env(_home_team(2) + _away_team(2), initializer=initializer)
    root_bodies = [p.walker.root_body for p in env.task.players]
    xy_bounds = np.asarray(env.task.arena.size) * spawn_ratio
    env.reset()
    xy = env.physics.bind(root_bodies).xpos[:, :2].copy()
    with self.subTest("X and Y positions within bounds"):
      if np.any(abs(xy) > xy_bounds):
        self.fail("Walker(s) spawned out of bounds. Expected abs(xy) "
                  "<= {}, got:\n{}".format(xy_bounds, xy))
    env.reset()
    xy2 = env.physics.bind(root_bodies).xpos[:, :2].copy()
    with self.subTest("X and Y positions change after reset"):
      if np.any(xy == xy2):
        self.fail("Walker(s) have the same X and/or Y coordinates before and "
                  "after reset. Before: {}, after: {}.".format(xy, xy2))

  def test_walker_rotation(self):
    initializer = initializers.UniformInitializer()
    env = _env(_home_team(2) + _away_team(2), initializer=initializer)

    def quats_to_eulers(quats):
      eulers = np.empty((len(quats), 3), dtype=np.double)
      dt = 1.
      for i, quat in enumerate(quats):
        mjbindings.mjlib.mju_quat2Vel(eulers[i], quat, dt)
      return eulers

    # TODO(b/132671988): Switch to using `get_pose` to get the quaternion once
    #                    `BoxHead.get_pose` and `BoxHead.set_pose` are
    #                    implemented in a consistent way.
    def get_quat(walker):
      return env.physics.bind(walker.root_body).xquat

    env.reset()
    quats = [get_quat(p.walker) for p in env.task.players]
    eulers = quats_to_eulers(quats)
    with self.subTest("Rotation is about the Z-axis only"):
      np.testing.assert_array_equal(eulers[:, :2], 0.)

    env.reset()
    quats2 = [get_quat(p.walker) for p in env.task.players]
    eulers2 = quats_to_eulers(quats2)
    with self.subTest("Rotation about Z changes after reset"):
      if np.any(eulers[:, 2] == eulers2[:, 2]):
        self.fail("Walker(s) have the same rotation about Z before and "
                  "after reset. Before: {}, after: {}."
                  .format(eulers[:, 2], eulers2[:, 2]))

  # TODO(b/132759890): Remove `expectedFailure` decorator once `set_velocity`
  #                    works correctly for the `BoxHead` walker.
  @unittest.expectedFailure
  def test_walker_velocity(self):
    initializer = initializers.UniformInitializer()
    env = _env(_home_team(2) + _away_team(2), initializer=initializer)
    root_joints = []
    non_root_joints = []
    for player in env.task.players:
      attachment_frame = mjcf.get_attachment_frame(player.walker.mjcf_model)
      root_joints.extend(
          attachment_frame.find_all("joint", immediate_children_only=True))
      non_root_joints.extend(player.walker.mjcf_model.find_all("joint"))
    # Assign a non-zero sentinel value to the velocities of all root and
    # non-root joints.
    sentinel_velocity = 3.14
    env.physics.bind(root_joints + non_root_joints).qvel = sentinel_velocity
    # The initializer should zero the velocities of the root joints, but not the
    # non-root joints.
    initializer(env.task, env.physics, env.random_state)
    np.testing.assert_array_equal(env.physics.bind(non_root_joints).qvel,
                                  sentinel_velocity)
    np.testing.assert_array_equal(env.physics.bind(root_joints).qvel, 0.)

  @parameterized.parameters([
      dict(spawn_ratio=0.3, init_ball_z=0.4),
      dict(spawn_ratio=0.5, init_ball_z=0.6),
  ])
  def test_ball_position(self, spawn_ratio, init_ball_z):
    initializer = initializers.UniformInitializer(
        spawn_ratio=spawn_ratio, init_ball_z=init_ball_z)
    env = _env(_home_team(2) + _away_team(2), initializer=initializer)
    xy_bounds = np.asarray(env.task.arena.size) * spawn_ratio
    env.reset()
    position, _ = env.task.ball.get_pose(env.physics)
    xyz = position.copy()
    with self.subTest("X and Y positions within bounds"):
      if np.any(abs(xyz[:2]) > xy_bounds):
        self.fail("Ball spawned out of bounds. Expected abs(xy) "
                  "<= {}, got:\n{}".format(xy_bounds, xyz[:2]))
    with self.subTest("Z position equal to `init_ball_z`"):
      self.assertEqual(xyz[2], init_ball_z)
    env.reset()
    position, _ = env.task.ball.get_pose(env.physics)
    xyz2 = position.copy()
    with self.subTest("X and Y positions change after reset"):
      if np.any(xyz[:2] == xyz2[:2]):
        self.fail("Ball has the same XY position before and after reset. "
                  "Before: {}, after: {}.".format(xyz[:2], xyz2[:2]))

  def test_ball_velocity(self):
    initializer = initializers.UniformInitializer()
    env = _env(_home_team(1) + _away_team(1), initializer=initializer)
    ball_root_joint = mjcf.get_frame_freejoint(env.task.ball.mjcf_model)
    # Set the velocities of the ball root joint to a non-zero sentinel value.
    env.physics.bind(ball_root_joint).qvel = 3.14
    initializer(env.task, env.physics, env.random_state)
    # The initializer should set the ball velocity to zero.
    ball_velocity = env.physics.bind(ball_root_joint).qvel
    np.testing.assert_array_equal(ball_velocity, 0.)

if __name__ == "__main__":
  absltest.main()



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for OSMesaContext."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

# Internal dependencies.
from absl.testing import absltest
from dm_control import _render
import mock
from OpenGL import GL

MAX_WIDTH = 640
MAX_HEIGHT = 480

CONTEXT_PATH = _render.__name__ + '.pyopengl.osmesa_renderer.osmesa'
GL_ARRAYS_PATH = _render.__name__ + '.pyopengl.osmesa_renderer.arrays'


@unittest.skipUnless(
    _render.BACKEND == _render.constants.OSMESA,
    reason='OSMesa backend not selected.')
class OSMesaContextTest(absltest.TestCase):

  def test_init(self):
    mock_context = mock.MagicMock()
    with mock.patch(CONTEXT_PATH) as mock_osmesa:
      mock_osmesa.OSMesaCreateContextExt.return_value = mock_context
      renderer = _render.Renderer(MAX_WIDTH, MAX_HEIGHT)
      self.assertIs(renderer._context, mock_context)
      renderer.free()

  def test_make_current(self):
    mock_context = mock.MagicMock()
    mock_buffer = mock.MagicMock()
    with mock.patch(CONTEXT_PATH) as mock_osmesa:
      with mock.patch(GL_ARRAYS_PATH) as mock_glarrays:
        mock_osmesa.OSMesaCreateContextExt.return_value = mock_context
        mock_glarrays.GLfloatArray.zeros.return_value = mock_buffer
        renderer = _render.Renderer(MAX_WIDTH, MAX_HEIGHT)
        with renderer.make_current():
          pass
        mock_osmesa.OSMesaMakeCurrent.assert_called_once_with(
            mock_context, mock_buffer, GL.GL_FLOAT, MAX_WIDTH, MAX_HEIGHT)
        renderer.free()

  def test_freeing(self):
    mock_context = mock.MagicMock()
    with mock.patch(CONTEXT_PATH) as mock_osmesa:
      mock_osmesa.OSMesaCreateContextExt.return_value = mock_context
      renderer = _render.Renderer(MAX_WIDTH, MAX_HEIGHT)
      renderer.free()
      mock_osmesa.OSMesaDestroyContext.assert_called_once_with(mock_context)
      self.assertIsNone(renderer._context)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Soccer task episode initializers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc

import numpy as np
import six


_INIT_BALL_Z = 0.5
_SPAWN_RATIO = 0.6


@six.add_metaclass(abc.ABCMeta)
class Initializer(object):

  @abc.abstractmethod
  def __call__(self, task, physics, random_state):
    """Initialize episode for a task."""


class UniformInitializer(Initializer):
  """Uniformly initialize walkers and soccer ball over spawn_range."""

  def __init__(self,
               spawn_ratio=_SPAWN_RATIO,
               init_ball_z=_INIT_BALL_Z,
               max_collision_avoidance_retries=100):
    self._spawn_ratio = spawn_ratio
    self._init_ball_z = init_ball_z

    # Lazily initialize geom ids for contact avoidance.
    self._ball_geom_ids = None
    self._walker_geom_ids = None
    self._all_geom_ids = None
    self._max_retries = max_collision_avoidance_retries

  def _initialize_ball(self, ball, spawn_range, physics, random_state):
    x, y = random_state.uniform(-spawn_range, spawn_range)
    ball.set_pose(physics, [x, y, self._init_ball_z])
    # Note: this method is not always called immediately after `physics.reset()`
    #       so we need to explicitly zero out the velocity.
    ball.set_velocity(physics, velocity=0., angular_velocity=0.)

  def _initialize_walker(self, walker, spawn_range, physics, random_state):
    """Uniformly initialize walker in spawn_range."""
    walker.reinitialize_pose(physics, random_state)
    x, y = random_state.uniform(-spawn_range, spawn_range)
    (_, _, z), quat = walker.get_pose(physics)
    walker.set_pose(physics, [x, y, z], quat)
    rotation = random_state.uniform(-np.pi, np.pi)
    quat = [np.cos(rotation / 2), 0, 0, np.sin(rotation / 2)]
    walker.shift_pose(physics, quaternion=quat)
    # Note: this method is not always called immediately after `physics.reset()`
    #       so we need to explicitly zero out the velocity.
    walker.set_velocity(physics, velocity=0., angular_velocity=0.)

  def _initialize_entities(self, task, physics, random_state):
    spawn_range = np.asarray(task.arena.size) * self._spawn_ratio
    self._initialize_ball(task.ball, spawn_range, physics, random_state)
    for player in task.players:
      self._initialize_walker(player.walker, spawn_range, physics, random_state)

  def _initialize_geom_ids(self, task, physics):
    self._ball_geom_ids = {physics.bind(task.ball.geom)}
    self._walker_geom_ids = []
    for player in task.players:
      walker_geoms = player.walker.mjcf_model.find_all('geom')
      self._walker_geom_ids.append(set(physics.bind(walker_geoms).element_id))

    self._all_geom_ids = set(self._ball_geom_ids)
    for walker_geom_ids in self._walker_geom_ids:
      self._all_geom_ids |= walker_geom_ids

  def _has_relevant_contact(self, contact, geom_ids):
    other_geom_ids = self._all_geom_ids - geom_ids
    if ((contact.geom1 in geom_ids and contact.geom2 in other_geom_ids) or
        (contact.geom2 in geom_ids and contact.geom1 in other_geom_ids)):
      return True
    return False

  def __call__(self, task, physics, random_state):
    # Initialize geom_ids for collision detection.
    if not self._all_geom_ids:
      self._initialize_geom_ids(task, physics)

    num_retries = 0
    while True:
      self._initialize_entities(task, physics, random_state)

      should_retry = False
      physics.forward()  # forward physics for contact resolution.
      for contact in physics.data.contact:
        if self._has_relevant_contact(contact, self._ball_geom_ids):
          should_retry = True
          break
        for walker_geom_ids in self._walker_geom_ids:
          if self._has_relevant_contact(contact, walker_geom_ids):
            should_retry = True
            break

      if not should_retry:
        break

      num_retries += 1
      if num_retries > self._max_retries:
        raise RuntimeError('UniformInitializer: `max_retries` (%d) exceeded.' %
                           self._max_retries)



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Suite-specific arena class."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control import composer


class Standard(composer.Arena):
  """Suite-specific subclass of the standard Composer arena."""

  def _build(self, name=None):
    """Initializes this arena.

    Args:
      name: (optional) A string, the name of this arena. If `None`, use the
        model name defined in the MJCF file.
    """
    super(Standard, self)._build(name=name)

    # Add visual assets.
    self.mjcf_model.asset.add(
        'texture',
        type='skybox',
        builtin='gradient',
        rgb1=(0.4, 0.6, 0.8),
        rgb2=(0., 0., 0.),
        width=100,
        height=100)
    groundplane_texture = self.mjcf_model.asset.add(
        'texture',
        name='groundplane',
        type='2d',
        builtin='checker',
        rgb1=(0.2, 0.3, 0.4),
        rgb2=(0.1, 0.2, 0.3),
        width=300,
        height=300,
        mark='edge',
        markrgb=(.8, .8, .8))
    groundplane_material = self.mjcf_model.asset.add(
        'material',
        name='groundplane',
        texture=groundplane_texture,
        texrepeat=(5, 5),
        texuniform='true',
        reflectance=0.2)

    # Add ground plane.
    self.mjcf_model.worldbody.add(
        'geom',
        name='ground',
        type='plane',
        material=groundplane_material,
        size=(1, 1, 0.1),
        friction=(0.4,),
        solimp=(0.95, 0.99, 0.001),
        solref=(0.002, 1))

    # Add lighting
    self.mjcf_model.worldbody.add(
        'light',
        pos=(0, 0, 1.5),
        dir=(0, 0, -1),
        diffuse=(0.7, 0.7, 0.7),
        specular=(.3, .3, .3),
        directional='false',
        castshadow='true')

    # Always initialize the free camera so that it points at the origin.
    self.mjcf_model.statistic.center = (0., 0., 0.)

  def attach_offset(self, entity, offset, attach_site=None):
    """Attaches another entity at a position offset from the attachment site.

    Args:
      entity: The `Entity` to attach.
      offset: A length 3 array-like object representing the XYZ offset.
      attach_site: (optional) The site to which to attach the entity's model.
        If not set, defaults to self.attachment_site.
    Returns:
      The frame of the attached model.
    """
    frame = self.attach(entity, attach_site=attach_site)
    frame.pos = offset
    return frame



# Copyright 2017-2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""OpenGL context management for rendering MuJoCo scenes.

By default, the `Renderer` class will try to load one of the following rendering
APIs, in descending order of priority: GLFW > EGL > OSMesa.

It is also possible to select a specific backend by setting the `MUJOCO_GL=`
environment variable to 'glfw', 'egl', or 'osmesa'.
"""

import collections
import os

from absl import logging
from dm_control._render import constants

BACKEND = os.environ.get(constants.MUJOCO_GL)


# pylint: disable=g-import-not-at-top
def _import_egl():
  from dm_control._render.pyopengl.egl_renderer import EGLContext
  return EGLContext


def _import_glfw():
  from dm_control._render.glfw_renderer import GLFWContext
  return GLFWContext


def _import_osmesa():
  from dm_control._render.pyopengl.osmesa_renderer import OSMesaContext
  return OSMesaContext
# pylint: enable=g-import-not-at-top

_ALL_RENDERERS = collections.OrderedDict([
    (constants.GLFW, _import_glfw),
    (constants.EGL, _import_egl),
    (constants.OSMESA, _import_osmesa),
])


if BACKEND is not None:
  # If a backend was specified, try importing it and error if unsuccessful.
  try:
    import_func = _ALL_RENDERERS[BACKEND]
  except KeyError:
    raise RuntimeError(
        'Environment variable {} must be one of {!r}: got {!r}.'
        .format(constants.MUJOCO_GL, _ALL_RENDERERS.keys(), BACKEND))
  logging.info('MUJOCO_GL=%s, attempting to import specified OpenGL backend.',
               BACKEND)
  Renderer = import_func()  # pylint: disable=invalid-name
else:
  logging.info('MUJOCO_GL is not set, so an OpenGL backend will be chosen '
               'automatically.')
  # Otherwise try importing them in descending order of priority until
  # successful.
  for name, import_func in _ALL_RENDERERS.items():
    try:
      Renderer = import_func()
      BACKEND = name
      logging.info('Successfully imported OpenGL backend: %s', name)
      break
    except ImportError:
      logging.info('Failed to import OpenGL backend: %s', name)
  if BACKEND is None:
    logging.info('No OpenGL backend could be imported. Attempting to create a '
                 'rendering context will result in a RuntimeError.')

    def Renderer(*args, **kwargs):  # pylint: disable=function-redefined,invalid-name
      del args, kwargs
      raise RuntimeError('No OpenGL rendering backend is available.')

USING_GPU = BACKEND in (constants.EGL, constants.GLFW)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

r"""Optimal policy for LQR levels.

LQR control problem is described in
https://en.wikipedia.org/wiki/Linear-quadratic_regulator#Infinite-horizon.2C_discrete-time_LQR
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control.mujoco import wrapper
import numpy as np
import scipy.linalg as scipy_linalg


def solve(env):
  """Returns the optimal value and policy for LQR problem.

  Args:
    env: An instance of `control.EnvironmentV2` with LQR level.

  Returns:
    p: A numpy array, the Hessian of the optimal total cost-to-go (value
      function at state x) is V(x) = .5 * x' * p * x.
    k: A numpy array which gives the optimal linear policy u = k * x.
    beta: The maximum eigenvalue of (a + b * k). Under optimal policy, at
      timestep n the state tends to 0 like beta^n.

  Raises:
    RuntimeError: If the controlled system is unstable.
  """
  n = env.physics.model.nq  # number of DoFs
  m = env.physics.model.nu  # number of controls

  # Compute the mass matrix.
  mass = np.zeros((n, n))
  wrapper.mjbindings.mjlib.mj_fullM(env.physics.model.ptr, mass,
                                    env.physics.data.qM)

  # Compute input matrices a, b, q and r to the DARE solvers.
  # State transition matrix a.
  stiffness = np.diag(env.physics.model.jnt_stiffness.ravel())
  damping = np.diag(env.physics.model.dof_damping.ravel())
  dt = env.physics.model.opt.timestep

  j = np.linalg.solve(-mass, np.hstack((stiffness, damping)))
  a = np.eye(2 * n) + dt * np.vstack(
      (dt * j + np.hstack((np.zeros((n, n)), np.eye(n))), j))

  # Control transition matrix b.
  b = env.physics.data.actuator_moment.T
  bc = np.linalg.solve(mass, b)
  b = dt * np.vstack((dt * bc, bc))

  # State cost Hessian q.
  q = np.diag(np.hstack([np.ones(n), np.zeros(n)]))

  # Control cost Hessian r.
  r = env.task.control_cost_coef * np.eye(m)

  # Solve the discrete algebraic Riccati equation.
  p = scipy_linalg.solve_discrete_are(a, b, q, r)
  k = -np.linalg.solve(b.T.dot(p.dot(b)) + r, b.T.dot(p.dot(a)))

  # Under optimal policy, state tends to 0 like beta^n_timesteps
  beta = np.abs(np.linalg.eigvals(a + b.dot(k))).max()
  if beta >= 1.0:
    raise RuntimeError('Controlled system is unstable.')
  return p, k, beta



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Functions to manage the common assets for domains."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from dm_control.utils import io as resources

_SUITE_DIR = os.path.dirname(os.path.dirname(__file__))
_FILENAMES = [
    "./common/materials.xml",
    "./common/skybox.xml",
    "./common/visual.xml",
]

ASSETS = {filename: resources.GetResource(os.path.join(_SUITE_DIR, filename))
          for filename in _FILENAMES}


def read_model(model_filename):
  """Reads a model XML file and returns its contents as a string."""
  return resources.GetResource(os.path.join(_SUITE_DIR, model_filename))



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for observation.observation_updater."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import itertools
import math

# Internal dependencies.

from absl.testing import absltest
from absl.testing import parameterized
from dm_control.composer.observation import fake_physics
from dm_control.composer.observation import observable
from dm_control.composer.observation import updater
from dm_env import specs
import numpy as np
import six
from six.moves import range


class DeterministicSequence(object):

  def __init__(self, sequence):
    self._iter = itertools.cycle(sequence)

  def __call__(self, random_state=None):
    del random_state  # unused
    return six.next(self._iter)


class BoundedGeneric(observable.Generic):

  def __init__(self, raw_observation_callable, minimum, maximum, **kwargs):
    super(BoundedGeneric, self).__init__(
        raw_observation_callable=raw_observation_callable,
        **kwargs)
    self._bounds = (minimum, maximum)

  @property
  def array_spec(self):
    datum = np.array(self(None, None))
    return specs.BoundedArray(shape=datum.shape,
                              dtype=datum.dtype,
                              minimum=self._bounds[0],
                              maximum=self._bounds[1])


class UpdaterTest(parameterized.TestCase):

  @parameterized.parameters(list, tuple)
  def testNestedSpecsAndValues(self, list_or_tuple):
    observables = list_or_tuple((
        {'one': observable.Generic(lambda _: 1.),
         'two': observable.Generic(lambda _: [2, 2]),
        }, collections.OrderedDict([
            ('three', observable.Generic(lambda _: np.full((2, 2), 3))),
            ('four', observable.Generic(lambda _: [4.])),
            ('five', observable.Generic(lambda _: 5)),
            ('six', BoundedGeneric(lambda _: [2, 2], 1, 4)),
            ('seven', BoundedGeneric(lambda _: 2, 1, 4, aggregator='sum')),
        ])
    ))

    observables[0]['two'].enabled = True
    observables[1]['three'].enabled = True
    observables[1]['five'].enabled = True
    observables[1]['six'].enabled = True
    observables[1]['seven'].enabled = True

    observation_updater = updater.Updater(observables)
    observation_updater.reset(physics=fake_physics.FakePhysics(),
                              random_state=None)

    def make_spec(obs):
      array = np.array(obs.observation_callable(None, None)())
      shape = array.shape if obs.aggregator else (1,) + array.shape

      if (isinstance(obs, BoundedGeneric) and
          obs.aggregator is not observable.base.AGGREGATORS['sum']):
        return specs.BoundedArray(shape=shape,
                                  dtype=array.dtype,
                                  minimum=obs.array_spec.minimum,
                                  maximum=obs.array_spec.maximum)
      else:
        return specs.Array(shape=shape, dtype=array.dtype)

    expected_specs = list_or_tuple((
        {'two': make_spec(observables[0]['two'])},
        collections.OrderedDict([
            ('three', make_spec(observables[1]['three'])),
            ('five', make_spec(observables[1]['five'])),
            ('six', make_spec(observables[1]['six'])),
            ('seven', make_spec(observables[1]['seven'])),
        ])
    ))

    actual_specs = observation_updater.observation_spec()
    self.assertIs(type(actual_specs), type(expected_specs))
    for actual_dict, expected_dict in zip(actual_specs, expected_specs):
      self.assertIs(type(actual_dict), type(expected_dict))
      self.assertEqual(actual_dict, expected_dict)

    def make_value(obs):
      value = obs(physics=None, random_state=None)
      if obs.aggregator:
        return value
      else:
        value = np.array(value)
        value = value[np.newaxis, ...]
        return value

    expected_values = list_or_tuple((
        {'two': make_value(observables[0]['two'])},
        collections.OrderedDict([
            ('three', make_value(observables[1]['three'])),
            ('five', make_value(observables[1]['five'])),
            ('six', make_value(observables[1]['six'])),
            ('seven', make_value(observables[1]['seven'])),
        ])
    ))

    actual_values = observation_updater.get_observation()
    self.assertIs(type(actual_values), type(expected_values))
    for actual_dict, expected_dict in zip(actual_values, expected_values):
      self.assertIs(type(actual_dict), type(expected_dict))
      self.assertLen(actual_dict, len(expected_dict))
      for actual, expected in zip(six.iteritems(actual_dict),
                                  six.iteritems(expected_dict)):
        actual_name, actual_value = actual
        expected_name, expected_value = expected
        self.assertEqual(actual_name, expected_name)
        np.testing.assert_array_equal(actual_value, expected_value)

  def assertCorrectSpec(
      self, spec, expected_shape, expected_dtype, expected_name):
    self.assertEqual(spec.shape, expected_shape)
    self.assertEqual(spec.dtype, expected_dtype)
    self.assertEqual(spec.name, expected_name)

  def testObservationSpecInference(self):
    physics = fake_physics.FakePhysics()
    physics.observables['repeated'].buffer_size = 5
    physics.observables['matrix'].buffer_size = 4
    physics.observables['sqrt'] = observable.Generic(
        fake_physics.FakePhysics.sqrt, buffer_size=3)

    for obs in six.itervalues(physics.observables):
      obs.enabled = True

    observation_updater = updater.Updater(physics.observables)
    observation_updater.reset(physics=physics, random_state=None)

    spec = observation_updater.observation_spec()
    self.assertCorrectSpec(spec['repeated'], (5, 2), np.int, 'repeated')
    self.assertCorrectSpec(spec['matrix'], (4, 2, 3), np.int, 'matrix')
    self.assertCorrectSpec(spec['sqrt'], (3,), np.float, 'sqrt')

  def testObservation(self):
    physics = fake_physics.FakePhysics()
    physics.observables['repeated'].buffer_size = 5
    physics.observables['matrix'].delay = 1
    physics.observables['sqrt'] = observable.Generic(
        fake_physics.FakePhysics.sqrt, update_interval=7,
        buffer_size=3, delay=2)
    for obs in six.itervalues(physics.observables):
      obs.enabled = True
    with physics.reset_context():
      pass

    physics_steps_per_control_step = 5
    observation_updater = updater.Updater(
        physics.observables, physics_steps_per_control_step)
    observation_updater.reset(physics=physics, random_state=None)

    for control_step in range(0, 200):
      observation_updater.prepare_for_next_control_step()
      for _ in range(physics_steps_per_control_step):
        physics.step()
        observation_updater.update()

      step_counter = (control_step + 1) * physics_steps_per_control_step

      observation = observation_updater.get_observation()
      def assert_correct_buffer(obs_name, expected_callable,
                                observation=observation,
                                step_counter=step_counter):
        update_interval = (physics.observables[obs_name].update_interval
                           or updater.DEFAULT_UPDATE_INTERVAL)
        buffer_size = (physics.observables[obs_name].buffer_size
                       or updater.DEFAULT_BUFFER_SIZE)
        delay = (physics.observables[obs_name].delay
                 or updater.DEFAULT_DELAY)

        # The final item in the buffer is the current time, less the delay,
        # rounded _down_ to the nearest multiple of the update interval.
        end = update_interval * int(
            math.floor((step_counter - delay) / update_interval))

        # Figure out the first item in the buffer by working backwards from
        # the final item in multiples of the update interval.
        start = end - (buffer_size - 1) * update_interval

        # Clamp both the start and end step number below by zero.
        buffer_range = range(max(0, start), max(0, end + 1), update_interval)

        # Arrays with expected shapes, filled with expected default values.
        expected_value_spec = observation_updater.observation_spec()[obs_name]
        expected_values = np.zeros(shape=expected_value_spec.shape,
                                   dtype=expected_value_spec.dtype)

        # The arrays are filled from right to left, such that the most recent
        # entry is the rightmost one, and any padding is on the left.
        for index, timestamp in enumerate(reversed(buffer_range)):
          expected_values[-(index+1)] = expected_callable(timestamp)

        np.testing.assert_array_equal(observation[obs_name], expected_values)

      assert_correct_buffer('twice', lambda x: 2*x)
      assert_correct_buffer('matrix', lambda x: [[x]*3]*2)
      assert_correct_buffer('repeated', lambda x: [x, x])
      assert_correct_buffer('sqrt', np.sqrt)

  def testVariableRatesAndDelays(self):
    physics = fake_physics.FakePhysics()
    physics.observables['time'] = observable.Generic(
        lambda physics: physics.time(),
        buffer_size=3,
        # observations produced on step numbers 20*N + [0, 3, 5, 8, 11, 15, 16]
        update_interval=DeterministicSequence([3, 2, 3, 3, 4, 1, 4]),
        # observations arrive on step numbers 20*N + [3, 8, 7, 12, 11, 17, 20]
        delay=DeterministicSequence([3, 5, 2, 5, 1, 2, 4]))
    physics.observables['time'].enabled = True

    physics_steps_per_control_step = 10
    observation_updater = updater.Updater(
        physics.observables, physics_steps_per_control_step)
    observation_updater.reset(physics=physics, random_state=None)

    # Run through a few cycles of the variation sequences to make sure that
    # cross-control-boundary behaviour is correct.
    for i in range(5):
      observation_updater.prepare_for_next_control_step()
      for _ in range(physics_steps_per_control_step):
        physics.step()
        observation_updater.update()
      np.testing.assert_array_equal(
          observation_updater.get_observation()['time'],
          20*i + np.array([0, 5, 3]))

      observation_updater.prepare_for_next_control_step()
      for _ in range(physics_steps_per_control_step):
        physics.step()
        observation_updater.update()
      # Note that #11 is dropped since it arrives after #8,
      # whose large delay caused it to cross the control step boundary at #10.
      np.testing.assert_array_equal(
          observation_updater.get_observation()['time'],
          20*i + np.array([8, 15, 16]))

if __name__ == '__main__':
  absltest.main()



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""LabMaze textures."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control import composer
from dm_control import mjcf
from labmaze import assets as labmaze_assets
import six


class SkyBox(composer.Entity):
  """Represents a texture asset for the sky box."""

  def _build(self, style):
    labmaze_textures = labmaze_assets.get_sky_texture_paths(style)
    self._mjcf_root = mjcf.RootElement(model='labmaze_' + style)
    self._texture = self._mjcf_root.asset.add(
        'texture', type='skybox', name='texture',
        fileleft=labmaze_textures.left, fileright=labmaze_textures.right,
        fileup=labmaze_textures.up, filedown=labmaze_textures.down,
        filefront=labmaze_textures.front, fileback=labmaze_textures.back)

  @property
  def mjcf_model(self):
    return self._mjcf_root

  @property
  def texture(self):
    return self._texture


class WallTextures(composer.Entity):
  """Represents wall texture assets."""

  def _build(self, style):
    labmaze_textures = labmaze_assets.get_wall_texture_paths(style)
    self._mjcf_root = mjcf.RootElement(model='labmaze_' + style)
    self._textures = []
    for texture_name, texture_path in six.iteritems(labmaze_textures):
      self._textures.append(self._mjcf_root.asset.add(
          'texture', type='2d', name=texture_name,
          file=texture_path.format(texture_name)))

  @property
  def mjcf_model(self):
    return self._mjcf_root

  @property
  def textures(self):
    return self._textures


class FloorTextures(composer.Entity):
  """Represents floor texture assets."""

  def _build(self, style):
    labmaze_textures = labmaze_assets.get_floor_texture_paths(style)
    self._mjcf_root = mjcf.RootElement(model='labmaze_' + style)
    self._textures = []
    for texture_name, texture_path in six.iteritems(labmaze_textures):
      self._textures.append(self._mjcf_root.asset.add(
          'texture', type='2d', name=texture_name,
          file=texture_path.format(texture_name)))

  @property
  def mjcf_model(self):
    return self._mjcf_root

  @property
  def textures(self):
    return self._textures



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Implements PyMJCF debug mode.

PyMJCF debug mode stores a stack trace each time the MJCF object is modified.
If Mujoco raises a compile error on the generated XML model, we would then be
able to find the original source line that created the offending element.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import contextlib
import copy
import os
import re
import sys
import traceback

from absl import flags
from lxml import etree
import six

FLAGS = flags.FLAGS
flags.DEFINE_boolean(
    'pymjcf_debug', False,
    'Enables PyMJCF debug mode (SLOW!). In this mode, a stack trace is logged '
    'each the MJCF object is modified. This may be helpful in locating the '
    'Python source line corresponding to a problematic element in the '
    'generated XML.')
flags.DEFINE_string(
    'pymjcf_debug_full_dump_dir', '',
    'Path to dump full debug info when Mujoco error is encountered.')

StackTraceEntry = collections.namedtuple(
    'StackTraceEntry', ('filename', 'line_number', 'function_name', 'text'))

ElementDebugInfo = collections.namedtuple(
    'ElementDebugInfo', ('element', 'init_stack', 'attribute_stacks'))

MODULE_PATH = os.path.dirname(sys.modules[__name__].__file__)
DEBUG_METADATA_PREFIX = 'pymjcfdebug'

_DEBUG_METADATA_TAG_PREFIX = '<!--' + DEBUG_METADATA_PREFIX
_DEBUG_METADATA_SEARCH_PATTERN = re.compile(
    r'<!--{}:(\d+)-->'.format(DEBUG_METADATA_PREFIX))

# Modified by `freeze_current_stack_trace`.
_CURRENT_FROZEN_STACK = None

# These globals will take their default values from the `--pymjcf_debug` and
# `--pymjcf_debug_full_dump_dir` flags respectively. We cannot use `FLAGS` as
# global variables because flag parsing might not have taken place (e.g. when
# running `nosetests`).
_DEBUG_MODE_ENABLED = None
_DEBUG_FULL_DUMP_DIR = None


def debug_mode():
  """Returns a boolean that indicates whether PyMJCF debug mode is enabled."""
  global _DEBUG_MODE_ENABLED
  if _DEBUG_MODE_ENABLED is None:
    if FLAGS.is_parsed():
      _DEBUG_MODE_ENABLED = FLAGS.pymjcf_debug
    else:
      _DEBUG_MODE_ENABLED = FLAGS['pymjcf_debug'].default
  return _DEBUG_MODE_ENABLED


def enable_debug_mode():
  """Enables PyMJCF debug mode."""
  global _DEBUG_MODE_ENABLED
  _DEBUG_MODE_ENABLED = True


def disable_debug_mode():
  """Disables PyMJCF debug mode."""
  global _DEBUG_MODE_ENABLED
  _DEBUG_MODE_ENABLED = False


def get_full_dump_dir():
  """Gets the directory to dump full debug info files."""
  global _DEBUG_FULL_DUMP_DIR
  if _DEBUG_FULL_DUMP_DIR is None:
    if FLAGS.is_parsed():
      _DEBUG_FULL_DUMP_DIR = FLAGS.pymjcf_debug_full_dump_dir
    else:
      _DEBUG_FULL_DUMP_DIR = FLAGS['pymjcf_debug_full_dump_dir'].default
  return _DEBUG_FULL_DUMP_DIR


def set_full_dump_dir(dump_path):
  """Sets the directory to dump full debug info files."""
  global _DEBUG_FULL_DUMP_DIR
  _DEBUG_FULL_DUMP_DIR = dump_path


def get_current_stack_trace():
  """Returns the stack trace of the current execution frame.

  Returns:
    A list of `StackTraceEntry` named tuples corresponding to the current stack
    trace of the process, truncated to immediately before entry into
    PyMJCF internal code.
  """
  if _CURRENT_FROZEN_STACK:
    return copy.deepcopy(_CURRENT_FROZEN_STACK)
  else:
    return _get_actual_current_stack_trace()


def _get_actual_current_stack_trace():
  """Returns the stack trace of the current execution frame.

  Returns:
    A list of `StackTraceEntry` named tuples corresponding to the current stack
    trace of the process, truncated to immediately before entry into
    PyMJCF internal code.
  """
  raw_stack = traceback.extract_stack()
  processed_stack = []
  for raw_stack_item in raw_stack:
    stack_item = StackTraceEntry(*raw_stack_item)
    if (stack_item.filename.startswith(MODULE_PATH)
        and not stack_item.filename.endswith('_test.py')):
      break
    else:
      processed_stack.append(stack_item)
  return processed_stack


@contextlib.contextmanager
def freeze_current_stack_trace():
  """A context manager that freezes the stack trace.

  AVOID USING THIS CONTEXT MANAGER OUTSIDE OF INTERNAL PYMJCF IMPLEMENTATION,
  AS IT REDUCES THE USEFULNESS OF DEBUG MODE.

  If PyMJCF debug mode is enabled, calls to `debugging.get_current_stack_trace`
  within this context will always return the stack trace from when this context
  was entered.

  The frozen stack is global to this debugging module. That is, if the context
  is entered while another one is still active, then the stack trace of the
  outermost one is returned.

  This context significantly speeds up bulk operations in debug mode, e.g.
  parsing an existing XML string or creating a deeply-nested element, as it
  prevents the same stack trace from being repeatedly constructed.

  Yields:
    `None`
  """
  global _CURRENT_FROZEN_STACK
  if debug_mode() and _CURRENT_FROZEN_STACK is None:
    _CURRENT_FROZEN_STACK = _get_actual_current_stack_trace()
    yield
    _CURRENT_FROZEN_STACK = None
  else:
    yield


class DebugContext(object):
  """A helper object to store debug information for a generated XML string.

  This class is intended for internal use within the PyMJCF implementation.
  """

  def __init__(self):
    self._xml_string = None
    self._debug_info_for_element_ids = {}

  def register_element_for_debugging(self, elem):
    """Registers an `Element` and returns debugging metadata for the XML.

    Args:
      elem: An `mjcf.Element`.

    Returns:
      An `lxml.etree.Comment` that represents debugging metadata in the
      generated XML.
    """
    if not debug_mode():
      return None
    else:
      self._debug_info_for_element_ids[id(elem)] = ElementDebugInfo(
          elem,
          copy.deepcopy(elem.get_init_stack()),
          copy.deepcopy(elem.get_last_modified_stacks_for_all_attributes()))
      return etree.Comment('{}:{}'.format(DEBUG_METADATA_PREFIX, id(elem)))

  def commit_xml_string(self, xml_string):
    """Commits the XML string associated with this debug context.

    This function also formats the XML string to make sure that the debugging
    metadata appears on the same line as the corresponding XML element.

    Args:
      xml_string: A pretty-printed XML string.

    Returns:
      A reformatted XML string where all debugging metadata appears on the same
      line as the corresponding XML element.
    """
    formatted = re.sub(r'\n\s*' + _DEBUG_METADATA_TAG_PREFIX,
                       _DEBUG_METADATA_TAG_PREFIX, xml_string)
    self._xml_string = formatted
    return formatted

  def process_and_raise_last_exception(self):
    """Processes and re-raises the last mujoco.wrapper.Error caught.

    This function will insert the relevant line from the source XML to the error
    message. If debug mode is enabled, additional debugging information is
    appended to the error message. If debug mode is not enabled, the error
    message instructs the user to enable it by rerunning the executable with an
    appropriate flag.
    """
    err_type, err, stack = sys.exc_info()
    line_number_match = re.search(r'[Ll][Ii][Nn][Ee]\s*[:=]?\s*(\d+)', str(err))
    if line_number_match:
      xml_line_number = int(line_number_match.group(1))
      xml_line = self._xml_string.split('\n')[xml_line_number - 1]
      stripped_xml_line = xml_line.strip()
      comment_match = re.search(_DEBUG_METADATA_TAG_PREFIX, stripped_xml_line)
      if comment_match:
        stripped_xml_line = stripped_xml_line[:comment_match.start()]
    else:
      xml_line = ''
      stripped_xml_line = ''

    message_lines = []
    if debug_mode():
      if get_full_dump_dir():
        self.dump_full_debug_info_to_disk()
      message_lines.extend([
          'Compile error raised by Mujoco.',
          str(err)])
      if xml_line:
        message_lines.extend([
            stripped_xml_line,
            self._generate_debug_message_from_xml_line(xml_line)])
    else:
      message_lines.extend([
          'Compile error raised by Mujoco; '
          'run again with --pymjcf_debug for additional debug information.',
          str(err)])
      if xml_line:
        message_lines.append(stripped_xml_line)

    message = '\n'.join(message_lines)
    six.reraise(err_type, err_type(message), stack)

  @property
  def default_dump_dir(self):
    return get_full_dump_dir()

  @property
  def debug_mode(self):
    return debug_mode()

  def dump_full_debug_info_to_disk(self, dump_dir=None):
    """Dumps full debug information to disk.

    Full debug information consists of an XML file whose elements are tagged
    with a unique ID, and a stack trace file for each element ID. Each stack
    trace file consists of a stack trace for when the element was created, and
    when each attribute was last modified.

    Args:
      dump_dir: Full path to the directory in which dump files are created.

    Raises:
      ValueError: If neither `dump_dir` nor the global dump path is given. The
        global dump path can be specified either via the
        --pymjcf_debug_full_dump_dir flag or via `debugging.set_full_dump_dir`.
    """
    dump_dir = dump_dir or self.default_dump_dir
    if not dump_dir:
      raise ValueError('`dump_dir` is not specified')
    section_separator = '\n' + ('=' * 80) + '\n'
    def dump_stack(header, stack, f):
      indent = '    '
      f.write(header + '\n')
      for stack_entry in stack:
        f.write(indent + '`{}` at {}:{}\n'
                .format(stack_entry.function_name,
                        stack_entry.filename, stack_entry.line_number))
        f.write((indent * 2) + str(stack_entry.text) + '\n')
      f.write(section_separator)
    with open(os.path.join(dump_dir, 'model.xml'), 'w') as f:
      f.write(self._xml_string)
    for elem_id, debug_info in six.iteritems(self._debug_info_for_element_ids):
      with open(os.path.join(dump_dir, str(elem_id) + '.dump'), 'w') as f:
        f.write('{}:{}\n'.format(DEBUG_METADATA_PREFIX, elem_id))
        f.write(str(debug_info.element) + '\n')
        dump_stack('Element creation', debug_info.init_stack, f)
        for attrib_name, stack in six.iteritems(debug_info.attribute_stacks):
          attrib_value = (
              debug_info.element.get_attribute_xml_string(attrib_name))
          if stack[-1] == debug_info.init_stack[-1]:
            if attrib_value is not None:
              f.write('Attribute {}="{}"\n'.format(attrib_name, attrib_value))
              f.write('    was set when the element was created\n')
              f.write(section_separator)
          else:
            if attrib_value is not None:
              dump_stack('Attribute {}="{}"'.format(attrib_name, attrib_value),
                         stack, f)
            else:
              dump_stack(
                  'Attribute {} was CLEARED'.format(attrib_name), stack, f)

  def _generate_debug_message_from_xml_line(self, xml_line):
    """Generates a debug message by parsing the metadata on an XML line."""
    metadata_match = _DEBUG_METADATA_SEARCH_PATTERN.search(xml_line)
    if metadata_match:
      elem_id = int(metadata_match.group(1))
      return self._generate_debug_message_from_element_id(elem_id)
    else:
      return ''

  def _generate_debug_message_from_element_id(self, elem_id):
    """Generates a debug message for the specified Element."""
    out = []
    debug_info = self._debug_info_for_element_ids[elem_id]

    out.append('Debug summary for element:')
    if not get_full_dump_dir():
      out.append(
          '  * Full debug info can be dumped to disk by setting the '
          'flag --pymjcf_debug_full_dump_dir=path/to/dump>')
    out.append('  * Element object was created by `{}` at {}:{}'
               .format(debug_info.init_stack[-1].function_name,
                       debug_info.init_stack[-1].filename,
                       debug_info.init_stack[-1].line_number))

    for attrib_name, stack in six.iteritems(debug_info.attribute_stacks):
      attrib_value = debug_info.element.get_attribute_xml_string(attrib_name)
      if stack[-1] == debug_info.init_stack[-1]:
        if attrib_value is not None:
          out.append('  * {}="{}" was set when the element was created'
                     .format(attrib_name, attrib_value))
      else:
        if attrib_value is not None:
          out.append('  * {}="{}" was set by `{}` at `{}:{}`'
                     .format(attrib_name, attrib_value,
                             stack[-1].function_name, stack[-1].filename,
                             stack[-1].line_number))
        else:
          out.append('  * {} was CLEARED by `{}` at {}:{}'
                     .format(attrib_name, stack[-1].function_name,
                             stack[-1].filename, stack[-1].line_number))

    return '\n'.join(out)



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.composer.environment."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.
from absl.testing import absltest
from absl.testing import parameterized
from dm_control import composer
from dm_control import mjcf
from dm_control.composer.observation import observable
import dm_env
import mock
import numpy as np
from six.moves import range


class DummyTask(composer.NullTask):

  def __init__(self):
    null_entity = composer.ModelWrapperEntity(mjcf.RootElement())
    super(DummyTask, self).__init__(null_entity)

  @property
  def task_observables(self):
    time = observable.Generic(lambda physics: physics.time())
    time.enabled = True
    return {'time': time}


class DummyTaskWithResetFailures(DummyTask):

  def __init__(self, num_reset_failures):
    super(DummyTaskWithResetFailures, self).__init__()
    self.num_reset_failures = num_reset_failures
    self.reset_counter = 0

  def initialize_episode_mjcf(self, random_state):
    self.reset_counter += 1

  def initialize_episode(self, physics, random_state):
    if self.reset_counter <= self.num_reset_failures:
      raise composer.EpisodeInitializationError()


class EnvironmentTest(parameterized.TestCase):

  def test_failed_resets(self):
    total_reset_failures = 5
    env_reset_attempts = 2
    task = DummyTaskWithResetFailures(num_reset_failures=total_reset_failures)
    env = composer.Environment(task, max_reset_attempts=env_reset_attempts)
    for _ in range(total_reset_failures // env_reset_attempts):
      with self.assertRaises(composer.EpisodeInitializationError):
        env.reset()
    env.reset()  # should not raise an exception
    self.assertEqual(task.reset_counter, total_reset_failures + 1)

  @parameterized.parameters(
      dict(name='reward_spec', defined_in_task=True),
      dict(name='reward_spec', defined_in_task=False),
      dict(name='discount_spec', defined_in_task=True),
      dict(name='discount_spec', defined_in_task=False))
  def test_get_spec(self, name, defined_in_task):
    task = DummyTask()
    env = composer.Environment(task)
    with mock.patch.object(task, 'get_' + name) as mock_task_get_spec:
      if defined_in_task:
        expected_spec = mock.Mock()
        mock_task_get_spec.return_value = expected_spec
      else:
        expected_spec = getattr(dm_env.Environment, name)(env)
        mock_task_get_spec.return_value = None
      spec = getattr(env, name)()
    mock_task_get_spec.assert_called_once_with()
    self.assertSameStructure(spec, expected_spec)

  def test_can_provide_observation(self):
    task = DummyTask()
    env = composer.Environment(task)
    obs = env.reset().observation
    self.assertLen(obs, 1)
    np.testing.assert_array_equal(obs['time'], env.physics.time())
    for _ in range(20):
      obs = env.step([]).observation
      self.assertLen(obs, 1)
      np.testing.assert_array_equal(obs['time'], env.physics.time())

if __name__ == '__main__':
  absltest.main()



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for the views.py module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from dm_control.viewer import views
import mock
import numpy as np
import six
from six.moves import range


class ColumnTextViewTest(absltest.TestCase):

  def setUp(self):
    super(ColumnTextViewTest, self).setUp()
    self.model = mock.MagicMock()
    self.view = views.ColumnTextView(self.model)

    self.context = mock.MagicMock()
    self.viewport = mock.MagicMock()
    self.location = views.PanelLocation.TOP_LEFT

  def test_rendering_empty_columns(self):
    self.model.get_columns.return_value = []
    with mock.patch(views.__name__ + '.mjlib') as mjlib_mock:
      self.view.render(self.context, self.viewport, self.location)
      self.assertEqual(0, mjlib_mock.mjr_overlay.call_count)

  def test_rendering(self):
    self.model.get_columns.return_value = [('', '')]
    with mock.patch(views.__name__ + '.mjlib') as mjlib_mock:
      self.view.render(self.context, self.viewport, self.location)
      mjlib_mock.mjr_overlay.assert_called_once()


class MujocoDepthBufferTests(absltest.TestCase):

  def setUp(self):
    super(MujocoDepthBufferTests, self).setUp()
    self.component = views.MujocoDepthBuffer()

    self.context = mock.MagicMock()
    self.viewport = mock.MagicMock()

  def test_updating_buffer_size_after_viewport_resize(self):
    self.component._depth_buffer = np.zeros((1, 1), np.float32)
    self.viewport.width = 10
    self.viewport.height = 10

    with mock.patch(views.__name__ + '.mjlib'):
      self.component.render(context=self.context, viewport=self.viewport)
      self.assertEqual((10, 10), self.component._depth_buffer.shape)

  def test_reading_depth_data(self):
    with mock.patch(views.__name__ + '.mjlib') as mjlib_mock:
      self.component.render(context=self.context, viewport=self.viewport)
      mjlib_mock.mjr_readPixels.assert_called_once()
      self.assertIsNone(mjlib_mock.mjr_readPixels.call_args[0][0])

  def test_rendering_position_fixed_to_bottom_right_quarter_of_viewport(self):
    self.viewport.width = 100
    self.viewport.height = 100
    expected_rect = [75, 0, 25, 25]
    with mock.patch(views.__name__ + '.mjlib') as mjlib_mock:
      self.component.render(context=self.context, viewport=self.viewport)
      mjlib_mock.mjr_drawPixels.assert_called_once()
      render_rect = mjlib_mock.mjr_drawPixels.call_args[0][2]
      self.assertEqual(expected_rect[0], render_rect.left)
      self.assertEqual(expected_rect[1], render_rect.bottom)
      self.assertEqual(expected_rect[2], render_rect.width)
      self.assertEqual(expected_rect[3], render_rect.height)


class ViewportLayoutTest(absltest.TestCase):

  def setUp(self):
    super(ViewportLayoutTest, self).setUp()
    self.layout = views.ViewportLayout()

    self.context = mock.MagicMock()
    self.viewport = mock.MagicMock()

  def test_added_elements_need_to_be_a_view(self):
    self.element = mock.MagicMock()
    with self.assertRaises(TypeError):
      self.layout.add(self.element, views.PanelLocation.TOP_LEFT)

  def test_adding_component(self):
    self.element = mock.MagicMock(spec=views.BaseViewportView)
    self.layout.add(self.element, views.PanelLocation.TOP_LEFT)
    self.assertLen(self.layout, 1)

  def test_adding_same_component_twice_updates_location(self):
    self.element = mock.MagicMock(spec=views.BaseViewportView)
    self.layout.add(self.element, views.PanelLocation.TOP_LEFT)
    self.layout.add(self.element, views.PanelLocation.TOP_RIGHT)
    self.assertEqual(
        views.PanelLocation.TOP_RIGHT, self.layout._views[self.element])

  def test_removing_component(self):
    self.element = mock.MagicMock(spec=views.BaseViewportView)
    self.layout._views[self.element] = views.PanelLocation.TOP_LEFT
    self.layout.remove(self.element)
    self.assertEmpty(self.layout)

  def test_removing_unregistered_component(self):
    self.element = mock.MagicMock(spec=views.BaseViewportView)
    self.layout.remove(self.element)  # No error is raised

  def test_clearing_layout(self):
    pos = views.PanelLocation.TOP_LEFT
    self.layout._views = {mock.MagicMock(spec=views.BaseViewportView): pos
                          for _ in range(3)}
    self.layout.clear()
    self.assertEmpty(self.layout)

  def test_rendering_layout(self):
    positions = [
        views.PanelLocation.TOP_LEFT,
        views.PanelLocation.TOP_RIGHT,
        views.PanelLocation.BOTTOM_LEFT]
    self.layout._views = {mock.MagicMock(spec=views.BaseViewportView): pos
                          for pos in positions}
    self.layout.render(self.context, self.viewport)
    for view, location in six.iteritems(self.layout._views):
      view.render.assert_called_once()
      self.assertEqual(location, view.render.call_args[0][2])


if __name__ == '__main__':
  absltest.main()



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Utilities and base classes used exclusively in the gui package."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import threading
import time

from dm_control.viewer import user_input
import six

_DOUBLE_CLICK_INTERVAL = 0.25  # seconds


@six.add_metaclass(abc.ABCMeta)
class InputEventsProcessor(object):
  """Thread safe input events processor."""

  def __init__(self):
    """Instance initializer."""
    self._lock = threading.RLock()
    self._events = []

  def add_event(self, receivers, *args):
    """Adds a new event to the processing queue."""
    if not all(callable(receiver) for receiver in receivers):
      raise TypeError('Receivers are expected to be callables.')
    def event():
      for receiver in list(receivers):
        receiver(*args)
    with self._lock:
      self._events.append(event)

  def process_events(self):
    """Invokes each of the events in the queue.

    Thread safe for queue access but not during event invocations.

    This method must be called regularly on the main thread.
    """
    with self._lock:
      # Swap event buffers quickly so that we don't block the input thread for
      # too long.
      events_to_process = self._events
      self._events = []

    # Now that we made the swap, process the received events in our own time.
    for event in events_to_process:
      event()


class DoubleClickDetector(object):
  """Detects double click events."""

  def __init__(self):
    self._double_clicks = {}

  def process(self, button, action):
    """Attempts to identify a mouse button click as a double click event."""
    if action != user_input.PRESS:
      return False

    curr_time = time.time()
    timestamp = self._double_clicks.get(button, None)
    if timestamp is None:
      # No previous click registered.
      self._double_clicks[button] = curr_time
      return False
    else:
      time_elapsed = curr_time - timestamp
      if time_elapsed < _DOUBLE_CLICK_INTERVAL:
        # Double click discovered.
        self._double_clicks[button] = None
        return True
      else:
        # The previous click was too long ago, so discard it and start a fresh
        # timer.
        self._double_clicks[button] = curr_time
        return False



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.locomotion.soccer.boxhead."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.

from absl.testing import absltest
from absl.testing import parameterized
from dm_control.locomotion.soccer import boxhead


class BoxheadTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(camera_control=True, walker_id=None),
      dict(camera_control=False, walker_id=None),
      dict(camera_control=True, walker_id=0),
      dict(camera_control=False, walker_id=10))
  def test_instantiation(self, camera_control, walker_id):
    boxhead.BoxHead(marker_rgba=[.8, .1, .1, 1.],
                    camera_control=camera_control,
                    walker_id=walker_id)

  @parameterized.parameters(-1, 11)
  def test_invalid_walker_id(self, walker_id):
    with self.assertRaisesWithLiteralMatch(
        ValueError, boxhead._INVALID_WALKER_ID.format(walker_id)):
      boxhead.BoxHead(walker_id=walker_id)


if __name__ == '__main__':
  absltest.main()
